# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoCard.ipynb.

# %% auto 0
__all__ = ['DomoCard', 'Card_DownloadSourceCode']

# %% ../../nbs/classes/50_DomoCard.ipynb 2
from domolibrary.client.DomoEntity import DomoEntity, DomoEntity_w_Lineage
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde

import domolibrary.routes.card as card_routes

import domolibrary.classes.DomoLineage as dmdl


import domolibrary.utils.DictDot as util_dd
import domolibrary.utils.chunk_execution as dmce
import domolibrary.utils.convert as dmut
import domolibrary.utils.files as dmfi

from dataclasses import dataclass, field
from typing import List, Any
from copy import deepcopy
import json


import httpx
from nbdev.showdoc import patch_to

# %% ../../nbs/classes/50_DomoCard.ipynb 5
@dataclass
class DomoCard(DomoEntity_w_Lineage):
    auth: dmda.DomoAuth = field(repr=False)
    id: str
    Lineage: "dmdl.DomoLineage" = field(repr=False)

    title: str = None
    description: str = None
    type: str = None
    urn: str = None
    chart_type: str = None
    dataset_id: str = None

    datastore_id: str = None

    domo_collections: List[Any] = None
    domo_source_code: Any = None

    certification: dict = None
    owners: List[any] = None

    datasets: List[Any] = field(repr=False, default=None)

    def __post_init__(self):
        # self.Definition = CardDefinition(self)
        self.Lineage = dmdl.DomoLineage._from_parent(auth=self.auth, parent=self)

    def display_url(self) -> str:
        return f"https://{self.auth.domo_instance}.domo.com/kpis/details/{self.id}"

    @classmethod
    async def _from_json(cls, card_obj, auth: dmda.DomoAuth):
        import domolibrary.classes.DomoUser as dmdu
        import domolibrary.classes.DomoGroup as dmgr

        dd = card_obj
        if isinstance(card_obj, dict):
            dd = util_dd.DictDot(card_obj)

        card = cls(
            auth=auth,
            id=dd.id,
            raw=card_obj,
            title=dd.title,
            description=dd.description,
            type=dd.type,
            urn=dd.urn,
            certification=dd.certification,
            chart_type=dd.metadata and dd.metadata.chartType,
            dataset_id=dd.datasources[0].dataSourceId if dd.datasources else None,
            Lineage=None,
        )

        tasks = []
        for user in dd.owners:
            if user.type == "USER":
                tasks.append(dmdu.DomoUser.get_by_id(user_id=user.id, auth=auth))
            if user.type == "GROUP":
                tasks.append(dmgr.DomoGroup.get_by_id(group_id=user.id, auth=auth))

        card.owners = await dmce.gather_with_concurrency(n=60, *tasks)

        if card_obj.get("domoapp", {}).get("id"):
            card.datastore_id = card_obj["domoapp"]["id"]

        return card

    @classmethod
    async def get_by_id(
        cls,
        card_id: str,
        auth: dmda.DomoAuth,
        optional_parts: str = "certification,datasources,drillPath,owners,properties,domoapp",
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        return_raw: bool = False,
    ):

        res = await card_routes.get_card_metadata(
            auth=auth,
            card_id=card_id,
            optional_parts=optional_parts,
            debug_api=debug_api,
            session=session,
            parent_class=cls.__name__,
        )

        if return_raw:
            return res

        domo_card = await cls._from_json(res.response, auth)

        return domo_card

    async def _get_entity_by_id(cls, entity_id: str, auth: dmda.DomoAuth, **kwargs):
        return await cls.get_by_id(card_id=entity_id, auth=auth, **kwargs)

# %% ../../nbs/classes/50_DomoCard.ipynb 7
@patch_to(DomoCard)
async def get_datasets(
    self,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):

    res = await card_routes.get_card_metadata(
        auth=self.auth,
        card_id=self.id,
        optional_parts="datasources",
        debug_api=debug_api,
        session=session,
        parent_class=self.__class__.__name__,
    )

    res.response = res.response["datasources"]

    if return_raw:
        return res

    import domolibrary.classes.DomoDataset as dmds

    self.datasets = await dmce.gather_with_concurrency(
        *[
            dmds.DomoDataset.get_by_id(dataset_id=obj["dataSourceId"], auth=self.auth)
            for obj in res.response
        ],
        n=10
    )

    return self.datasets

# %% ../../nbs/classes/50_DomoCard.ipynb 9
@patch_to(DomoCard)
async def share(
    self: DomoCard,
    auth: dmda.DomoAuth = None,
    domo_users: list = None,  # DomoUsers to share card with,
    domo_groups: list = None,  # DomoGroups to share card with
    message: str = None,  # message for automated email
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    import domolibrary.routes.datacenter as datacenter_routes

    if domo_groups:
        domo_groups = domo_groups if isinstance(domo_groups, list) else [domo_groups]
    if domo_users:
        domo_users = domo_users if isinstance(domo_users, list) else [domo_users]

    res = await datacenter_routes.share_resource(
        auth=auth or self.auth,
        resource_ids=[self.id],
        resource_type=datacenter_routes.ShareResource_Enum.CARD,
        group_ids=[group.id for group in domo_groups] if domo_groups else None,
        user_ids=[user.id for user in domo_users] if domo_users else None,
        message=message,
        debug_api=debug_api,
        session=session,
    )

    return res

# %% ../../nbs/classes/50_DomoCard.ipynb 12
class Card_DownloadSourceCode(dmde.DomoError):
    def __init__(self, cls, auth, message):
        super().__init__(cls=cls, auth=auth, message=message)


@patch_to(DomoCard)
async def get_collections(
    self, debug_api: bool = False, return_raw: bool = False, debug_num_stacks_to_drop=2
):
    import domolibrary.classes.DomoAppDb as dmdb

    domo_collections = await dmdb.AppDbCollections.get_collections(
        datastore_id=self.datastore_id,
        auth=self.auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        return_raw=return_raw,
    )

    if return_raw:
        return domo_collections

    self.domo_collections = await dmce.gather_with_concurrency(
        *[
            dmdb.AppDbCollection.get_by_id(
                collection_id=domo_collection.id, auth=self.auth, debug_api=debug_api
            )
            for domo_collection in domo_collections
        ],
        n=60,
    )

    return self.domo_collections


@patch_to(DomoCard)
async def get_source_code(self, debug_api: bool = False, try_auto_share: bool = False):

    await self.get_collections(debug_api=debug_api)

    collection_name = "ddx_app_client_code"
    code_collection = next(
        (
            domo_collection
            for domo_collection in self.domo_collections
            if domo_collection.name == collection_name
        ),
        None,
    )

    if not code_collection:
        raise Card_DownloadSourceCode(
            cls=deepcopy(self),
            auth=self.auth,
            message=f"collection - {collection_name} not found for {self.title} - {self.id}",
        )

    documents = await code_collection.query_documents(
        debug_api=debug_api, try_auto_share=try_auto_share
    )

    if not documents:
        raise Card_DownloadSourceCode(
            cls=deepcopy(self),
            auth=self.auth,
            message=f"collection - {collection_name} - {code_collection.id} - unable to retrieve documents for {self.title} - {self.id}",
        )

    self.domo_source_code = documents[0]

    return self.domo_source_code

# %% ../../nbs/classes/50_DomoCard.ipynb 14
@patch_to(DomoCard)
async def download_source_code(
    self,
    download_folder="./EXPORT/",
    file_name=None,
    debug_api: bool = False,
    try_auto_share: bool = False,
):
    doc = await self.get_source_code(debug_api=debug_api, try_auto_share=try_auto_share)

    if file_name:
        download_path = os.path.join(
            download_folder, dmut.change_suffix(file_name, new_extension=".json")
        )
        dmfi.upsert_folder(download_path)

        with open(download_path, "w+", encoding="utf-8") as f:
            f.write(json.dumps(doc.content))
            return doc

    ddx_type = next(iter(doc.content))

    for key, value in doc.content[ddx_type].items():
        if key == "js":
            file_name = "app.js"
        elif key == "html":
            file_name = "index.html"
        elif key == "css":
            file_name = "styles.css"
        else:
            file_name = f"{key}.txt"

        download_path = os.path.join(
            download_folder, f"{ddx_type}/{self.id}/{file_name}"
        )
        dmfi.upsert_folder(download_path)

        with open(download_path, "w+", encoding="utf-8") as f:
            f.write(value)

    return doc



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoAccessToken.ipynb.

# %% auto 0
__all__ = ['DomoAccessToken', 'DomoAccessTokens']

# %% ../../nbs/classes/50_DomoAccessToken.ipynb 3
import httpx
import datetime as dt
import asyncio
from typing import List, Any

from dataclasses import dataclass, field

import domolibrary.utils.convert as dmcv
import domolibrary.client.DomoAuth as dmda
import domolibrary.utils.chunk_execution as ce

import domolibrary.routes.access_token as access_token_routes

# %% ../../nbs/classes/50_DomoAccessToken.ipynb 7
@dataclass
class DomoAccessToken:
    id: int
    name: str
    owner: Any # DomoUser
    expiration_date: dt.datetime
    token: str = field(repr=False)

    auth: dmda.DomoAuth = field(repr=False)

    days_till_expiration : int = None


    def __eq__(self, other):
        if not isinstance(other, DomoAccessToken):
            return False
        return self.id == other.id

    def __post_init__(self):
        if not isinstance(self.expiration_date, dt.datetime):
            self.expiration_date = dmcv.convert_epoch_millisecond_to_datetime(self.expiration_date)

        self.calc_days_till_expiration()

    def calc_days_till_expiration(self):
        self.days_till_expiration = (self.expiration_date - dt.datetime.now()).days
        return self.days_till_expiration

    @classmethod
    async def _from_json(cls, obj, auth):
        import domolibrary.classes.DomoUser as dmu

        owner = await dmu.DomoUser.get_by_id(user_id=obj["ownerId"], auth=auth)

        return cls(
            id=obj["id"],
            name=obj["name"],
            owner=owner,
            expiration_date=obj["expires"],
            auth=auth,
            token=obj.get("token"),
        )
    
    @classmethod
    async def generate(
        cls,
        duration_in_days: int,
        token_name: str,
        auth: dmda.DomoAuth,
        owner,  # DomoUser
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
        parent_class: str = None,
    ):

        res = await access_token_routes.generate_access_token(
            user_id=owner.id,
            token_name=token_name,
            duration_in_days=duration_in_days,
            auth=auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=parent_class or cls.__name__,
        )

        if return_raw:
            return res

        return await cls._from_json(obj=res.response, auth=auth)

    async def revoke(
        self,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
    ):
        return await access_token_routes.revoke_access_token(
            auth=self.auth,
            access_token_id=self.id,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

    async def regenerate(
        self,
        session: httpx.AsyncClient = None,
        duration_in_days: int = 90,
        debug_api: bool = False,
        return_raw: bool = False,
        debug_num_stacks_to_drop: int = 2,
    ):

        await self.revoke()

        await asyncio.sleep(3)

        new_token = await self.generate(
            duration_in_days=duration_in_days,
            token_name=self.name,
            auth=self.auth,
            owner=self.owner,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            return_raw=return_raw,
            parent_class=self.__class__.__name__,
        )
        
        self.id = new_token.id
        self.token = new_token.token
        self.expiration_date = new_token.expiration_date

        return self

# %% ../../nbs/classes/50_DomoAccessToken.ipynb 10
@dataclass
class DomoAccessTokens:
    
    auth: dmda.DomoAuth = field(repr=False)

    domo_access_tokens: List[DomoAccessToken] = field(default=None)

    async def get(
        self,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        session: httpx.AsyncClient = None,
    ):
        res = await access_token_routes.get_access_tokens(
            auth=self.auth,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        return await ce.gather_with_concurrency(
            *[DomoAccessToken._from_json(obj=obj, auth=self.auth) for obj in res.response],
            n=10
        )
    

    async def generate(
        self,
        duration_in_days: int,
        token_name: str,
        owner,  # DomoUser
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop: int = 2,
        return_raw: bool = False,
    ):

        domo_access_token = await DomoAccessToken.generate(
            owner=owner,
            token_name=token_name,
            duration_in_days=duration_in_days,
            auth=self.auth,
            debug_api=debug_api,
            session=session,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            return_raw = return_raw
        )

        await self.get(session = session, debug_api = debug_api, debug_num_stacks_to_drop= debug_num_stacks_to_drop+1)

        return domo_access_token



"""preferred base error class for Domo Exceptions"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/client/DomoError.ipynb.

# %% auto 0
__all__ = ['DomoError', 'RouteError', 'ClassError']

# %% ../../nbs/client/DomoError.ipynb 2
from typing import Optional, Any
# import domolibrary.client.ResponseGetData as rgd

# %% ../../nbs/client/DomoError.ipynb 5
class DomoError(Exception):
    """base exception"""

    def __init__(
        self,
        entity_id: Optional[str] = None,
        function_name: Optional[str] = None,
        status: Optional[int] = None,  # API request status
        message: str = "error",  # <domo_instance>.domo.com
        res = None,
        domo_instance: Optional[str] = None,
        parent_class: str = None,
        is_exception_not_error: bool = False,
    ):
        function_str = ""

        if function_name:
            function_str = f"{function_name} || "

        if parent_class and function_name:
            function_str = f"{res and res.traceback_details.parent_class or parent_class}.{function_str}"
        if parent_class and not function_name:
            function_str = f"{res and res.traceback_details.parent_class or parent_class} || "
        function_str = f"function: {function_str}"

        entity_str = f"{entity_id} || " if entity_id else ""

        domo_instance = domo_instance or res and res.auth and res.auth.domo_instance or None
        instance_str = f" at {domo_instance }" if domo_instance else ""
        
        
        status_str = f"status {res and res.status or status} || " if status else ""
        prefix = "🛑 " if not is_exception_not_error else "⚠️ "

        message = f"{prefix} {self.__class__.__name__} {prefix}- {function_str}{entity_str}{status_str}{message}{instance_str}"
        super().__init__(message)

# %% ../../nbs/client/DomoError.ipynb 7
class RouteError(DomoError):
    """base exception"""

    def __init__(
        self,
        res : Any,
        entity_id: str = None,
        is_exception_not_error: bool = False,
        message=None,
        domo_instance : str = None, 
        **kwargs
    ):
        
        domo_instance = domo_instance or (res and res.auth and res.auth.domo_instance) or None

        err = {
            "entity_id": entity_id,
            "function_name": res.traceback_details.function_name if res and res.traceback_details else None,
            "status": res.status if res else None,
            "message": message or res.response if res else "error",
            "domo_instance": domo_instance,
            "parent_class": res.traceback_details.parent_class if res and res.traceback_details else None,
            "is_exception_not_error": is_exception_not_error,
            **kwargs
        }

        super().__init__(
            **err
        )

# %% ../../nbs/client/DomoError.ipynb 9
class ClassError(DomoError):
    """base exception"""

    def __init__(
        self,
        cls = None,
        cls_instance = None,
        cls_name_attr = None,
        entity_id: str = None,
        entity_name : str = None,
        res = None,
        auth = None,
        message :str = None
    ):
        cls = cls or cls_instance.__class__

        auth = auth or (res and res.auth)

        message_str = message or (res and res.response)

        function_str = f"{cls.__name__}"
        if res:
            if res.traceback_details.function_name:
                function_str = f"{function_str}.{res.traceback_details.function_name}"

        entity_id = entity_id or (cls_instance and cls_instance.id)
        entity_name = entity_name or (cls_instance and cls_name_attr and getattr(cls_instance, cls_name_attr)) or ""

        entity_str = entity_id

        if entity_name:
            entity_str = f"{entity_id} - {entity_name}" 
        
        instance_str = f" @ {(auth and auth.domo_instance)}" if auth else ""

        prefix = "🛑"

        message = f"{prefix} {function_str} {prefix} {entity_str} || {message_str}{instance_str}"
        super().__init__(message)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/account.ipynb.

# %% auto 0
__all__ = ['Account_GET_Error', 'get_available_data_providers', 'get_accounts', 'get_oauth_accounts', 'Account_NoMatch',
           'get_account_by_id', 'get_oauth_account_by_id', 'get_account_config', 'get_oauth_account_config',
           'Account_CRUD_Error', 'Account_CreateParams_Error', 'generate_create_account_body', 'create_account',
           'delete_account', 'generate_create_oauth_account_body', 'create_oauth_account', 'delete_oauth_account',
           'Account_Config_Error', 'update_account_config', 'update_oauth_account_config', 'update_account_name',
           'update_oauth_account_name', 'get_account_accesslist', 'get_oauth_account_accesslist', 'ShareAccount',
           'ShareAccount_V1_AccessLevel', 'ShareAccount_AccessLevel', 'Account_Share_Error',
           'Account_AlreadyShared_Error', 'share_account', 'share_oauth_account', 'share_account_v1']

# %% ../../nbs/routes/account.ipynb 3
from abc import abstractmethod
import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
from domolibrary.client.DomoEntity import DomoEnum

# %% ../../nbs/routes/account.ipynb 8
class Account_GET_Error(dmde.RouteError):
    def __init__(
        self, res: rgd.ResponseGetData, account_id: int = None, message: str = None
    ):
        super().__init__(res=res, message=message, entity_id=account_id)

@gd.route_function
async def get_available_data_providers(
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/providers"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise Account_GET_Error(res=res)
    return res



@gd.route_function
async def get_accounts(
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:
    """retrieve a list of all the accounts the user has read access to.  Note users with "Manage all accounts" will retrieve all account objects"""

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise Account_GET_Error(res=res)
    return res


@gd.route_function
async def get_oauth_accounts(
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/templates/user/extended"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise Account_GET_Error(res=res)
    return res

# %% ../../nbs/routes/account.ipynb 13
class Account_NoMatch(dmde.RouteError):
    def __init__(
        self,
        res: rgd.ResponseGetData,
        account_id=None,
        message=f"Account not found --  has it been shared with the user?",
    ):

        super().__init__(
            message=message,
            res=res,
            entity_id=account_id,
        )


@gd.route_function
async def get_account_by_id(
    auth: dmda.DomoAuth,
    account_id: int,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
    is_unmask: bool = False
) -> rgd.ResponseGetData:
    """retrieves metadata about an account"""

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        timeout=20,  # occasionally this API has a long response time
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        params = {"unmask" : is_unmask}
    )

    if not res.is_success and (
        res.response == "Forbidden" or res.response == "Not Found"
    ):
        raise Account_NoMatch(account_id=account_id, res=res)

    if not res.is_success:
        raise Account_GET_Error(res=res)

    return res


@gd.route_function
async def get_oauth_account_by_id(
    auth: dmda.DomoAuth,
    account_id: int,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """
    retrieves all_oauth_accounts and filters to the selected one.
    There does not appear to be an API implementation of retrieving one OAuth Account
    """

    res = await get_oauth_accounts(
        auth=auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        parent_class=parent_class,
        session=session,
    )

    if return_raw:
        return res

    res.response = next((obj for obj in res.response if obj["id"] == account_id), None)

    if not res.response:
        raise Account_NoMatch(res=res, account_id=account_id)

    return res

# %% ../../nbs/routes/account.ipynb 18
# class Account_NoConfigRetrieved(dmde.RouteError):
#     def __init__(
#         self,
#         res: rgd.ResponseGetData,
#         account_id=None,
#         message: str = f"account did not return a config.  update `DomoAccount_Config` if it uses OAuth, otherwise this is probably an error",
#     ):
#         super().__init__(message=message, res=res, entity_id=account_id)


@gd.route_function
async def get_account_config(
    auth: dmda.DomoAuth,
    account_id: int,
    data_provider_type: str = None,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
    is_unmask :bool = True # unmasks encrypted values IF config allows
) -> rgd.ResponseGetData:

    if not data_provider_type:
        res = await get_account_by_id(
            auth = auth,
            account_id=account_id,
            debug_api=debug_api,
            session=session,
            parent_class=parent_class,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        data_provider_type = res.response["dataProviderType"]

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/account/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        params = {"unmask" : is_unmask},
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res

    if not res.is_success and (
        res.response == "Forbidden" or res.response == "Not Found"
    ):
        raise Account_NoMatch(account_id=account_id, res=res)

    if not res.is_success:
        raise Account_GET_Error(res=res)

    res.response.update(
        {
            "_search_metadata": {
                "account_id": account_id,
                "data_provider_type": data_provider_type,
            }
        }
    )

    return res


@gd.route_function
async def get_oauth_account_config(
    auth: dmda.DomoAuth,
    account_id: int,
    data_provider_type: str,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:
    """retrieves metadata about an account"""

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/template/{account_id}?unmask=true"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        session=session,
        timeout=20,  # occasionally this API has a long response time
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res

    if not res.is_success and (
        res.response == "Forbidden" or res.response == "Not Found"
    ):
        raise Account_NoMatch(account_id=account_id, res=res)

    if not res.is_success:
        raise Account_GET_Error(res=res)

    res.response.update(
        {
            "_search_metadata": {
                "account_id": account_id,
                "data_provider_type": data_provider_type,
            }
        }
    )

    return res

# %% ../../nbs/routes/account.ipynb 23
class Account_CRUD_Error(dmde.RouteError):
    def __init__(
        self,
        res: rgd.ResponseGetData,
        account_id=None,
        message: str = None,
    ):
        message = message or f"unable to update account - {res.response}"

        super().__init__(message=message, entity_id=account_id, res=res)


class Account_CreateParams_Error(dmde.DomoError):
    def __init__(
        self,
        message: str,
    ):
        super().__init__(message=message)


def generate_create_account_body(account_name, data_provider_type, config_body):
    return {
        "displayName": account_name,
        "dataProviderType": data_provider_type,
        "name": account_name,
        "configurations": config_body,
    }

import json
@gd.route_function
async def create_account(
    auth: dmda.DomoAuth,
    account_name: str = None,
    data_provider_type: str = None,
    config_body: dict = None, # properly formatted AccountConfig 
    payload: dict = None, # final payload
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:

    
    payload = payload or generate_create_account_body(
        account_name=account_name,
        data_provider_type=data_provider_type,
        config_body=config_body,
    )
    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        body=payload,
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise Account_CRUD_Error(
            res=res,
            account_id=account_name,
            message=f"error creating account - {res.response}",
        )

    return res


@gd.route_function
async def delete_account(
    auth: dmda.DomoAuth,
    account_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="DELETE",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise Account_CRUD_Error(
            account_id=account_id,
            res=res,
            message=f"unable to delete account",
        )

    return res


def generate_create_oauth_account_body(
    account_name, data_provider_type, origin, config
):
    return {
        "name": account_name,
        "displayName": account_name,
        "dataProviderType": data_provider_type,
        "origin": origin,
        "configurations": config,
    }


@gd.route_function
async def create_oauth_account(
    auth: dmda.DomoAuth,
    account_name: str = None,
    data_provider_type: str = None,
    origin: str = "OAUTH_CONFIGURATION",
    config: dict = None,
    create_body: dict = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/templates"

    if not create_body and not (
        account_name and data_provider_type and origin and config
    ):
        raise Account_CreateParams_Error(
            message="if not passing complete create_body must pass account_name,data_provider_type, origin, and config"
        )

    create_body = create_body or generate_create_oauth_account_body(
        account_name=account_name,
        data_provider_type=data_provider_type,
        origin=origin,
        config=config,
    )

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        body=create_body,
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise Account_CRUD_Error(
            res=res,
            account_id=create_body.get("displayName"),
            message=f"error creating account - {res.response}",
        )

    return res


@gd.route_function
async def delete_oauth_account(
    auth: dmda.DomoAuth,
    account_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/templates/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="DELETE",
        debug_api=debug_api,
        session=session,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise Account_CRUD_Error(
            account_id=account_id, res=res, message="error deleting account"
        )

    res.response = f"deleted account {account_id}"
    return res

# %% ../../nbs/routes/account.ipynb 27
class Account_Config_Error(dmde.RouteError):
    def __init__(
        self,
        res: rgd.ResponseGetData,
        account_id=None,
        message: str = None,
    ):
        message = message or f"unable to update account - {res.response}"

        super().__init__(message=message, entity_id=account_id, res=res)


@gd.route_function
async def update_account_config(
    auth: dmda.DomoAuth,
    account_id: int,
    config_body: dict,
    data_provider_type: str = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:

    # get the data_provider_type, which is necessare for updating the config setting
    if not data_provider_type:
        res = await get_account_by_id(
            auth=auth,
            account_id=account_id,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=parent_class,
            session=session,
        )
        data_provider_type = res.response.get("dataProviderType")

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/account/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=config_body,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if res.status == 400 and res.response == "Bad Request":
        raise Account_Config_Error(
            res=res,
            message=f"error updating config | use debug_api = True {res.response}",
            account_id=account_id,
        )

    if not res.is_success:
        raise Account_CRUD_Error(res=res, account_id=account_id)

    return res

@gd.route_function
async def update_oauth_account_config(
    auth: dmda.DomoAuth,
    account_id: int,
    config_body: dict,
    data_provider_type: str = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:

    # get the data_provider_type, which is necessare for updating the config setting
    if not data_provider_type:
        res = await get_oauth_account_by_id(
            auth=auth,
            account_id=account_id,
            # debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=parent_class,
            session=session,
        )
        data_provider_type = res.response.get("dataProviderType")

    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/providers/{data_provider_type}/template/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=config_body,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if res.status == 400 and res.response == "Bad Request":
        raise Account_Config_Error(
            res=res,
            message=f"error updating config | use debug_api = True {res.response}",
            account_id=account_id,
        )

    if not res.is_success:
        raise Account_CRUD_Error(res=res, account_id=account_id)

    return res



# %% ../../nbs/routes/account.ipynb 31
@gd.route_function
async def update_account_name(
    auth: dmda.DomoAuth,
    account_id: int,
    account_name: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:
    url = (
        f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}/name"
    )

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=account_name,
        content_type="text/plain",
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )

    if not res.is_success:
        raise Account_CRUD_Error(
            res=res,
            account_id=account_id,
            message="error updating account_name",
        )

    return res


@gd.route_function
async def update_oauth_account_name(
    auth: dmda.DomoAuth,
    account_id: int,
    account_name: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:
    url = (
        f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/templates/{account_id}/name"
    )

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=account_name,
        content_type="text/plain",
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )

    if not res.is_success:
        raise Account_CRUD_Error(
            res=res,
            account_id=account_id,
            message="error updating account_name",
        )

    return res
    

# %% ../../nbs/routes/account.ipynb 36
@gd.route_function
async def get_account_accesslist(
    auth: dmda.DomoAuth,
    account_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
    return_raw: bool = False
):
    url = (
        f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/share/{account_id}"
    )

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise Account_GET_Error(
            res=res,
            account_id=account_id,
            message=f"unable to get accesslist - {res.response}",
        )

    if return_raw:
        return res
    
    res.response = res.response['list']

    return res


@gd.route_function
async def get_oauth_account_accesslist(
    auth: dmda.DomoAuth,
    account_id: str,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/templates/share/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
    )

    if not res.is_success:
        raise Account_GET_Error(
            res=res,
            account_id=account_id,
            message=f"unable to get accesslist - {res.response}",
        )

    if return_raw:
        return res
    
    res.response = res.response['list']

    return res

# %% ../../nbs/routes/account.ipynb 39
class ShareAccount(DomoEnum):
    @abstractmethod
    def generate_payload(self):
        pass


class ShareAccount_V1_AccessLevel(ShareAccount):
    CAN_VIEW = "READ"
    CAN_EDIT = "WRITE"
    OWNER = "OWNER"
    default = "READ"

    def generate_payload(self, user_id: int, **kwargs):

        return {"type": "USER", "id": int(user_id), "permissions": [self.value]}


class ShareAccount_AccessLevel(ShareAccount):
    """Enum that correlates to the v2 API which should also be the current version"""

    CAN_VIEW = "CAN_VIEW"
    CAN_EDIT = "CAN_EDIT"
    CAN_SHARE = "CAN_SHARE"
    OWNER = "OWNER"
    NO_ACCESS = "NONE"
    default = "CAN_VIEW"

    
    def generate_payload(self, user_id: int = None, group_id: int = None, **kwargs):
        if user_id:
            return {"type": "USER", "id": str(user_id), "accessLevel": self.value}

        if group_id:
            return {"type": "GROUP", "id": str(group_id), "accessLevel": self.value}

# %% ../../nbs/routes/account.ipynb 41
class Account_Share_Error(dmde.RouteError):
    def __init__(
        self, res: rgd.ResponseGetData, account_id: int = None, message: str = None
    ):
        super().__init__(res=res, message=message, entity_id=account_id)


class Account_AlreadyShared_Error(dmde.DomoError):
    def __init__(
        self, res: rgd.ResponseGetData, account_id: int = None, message: str = None
    ):
        super().__init__(res=res, message=message, entity_id=account_id)


@gd.route_function
async def share_account(
    auth: dmda.DomoAuth,
    account_id: str,
    share_payload: dict,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
    session: httpx.AsyncClient = None,
):
    """for simplicity, this is the v2 api which should have been deployed to all Domo instances as of DP24"""

    url = (
        f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/share/{account_id}"
    )

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=share_payload,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        debug_api=debug_api,
        session=session,
    )

    if res.status == 500 and res.response == "Internal Server Error":
        raise Account_AlreadyShared_Error(
            account_id=account_id,
            messagae=f"{res.response} - User may already have access to account",
            res=res,
        )

    if not res.is_success:
        raise Account_Share_Error(
            account_id=account_id,
            res=res,
        )

    return res


ShareAccount_AccessLevel.share = staticmethod(share_account)


@gd.route_function
async def share_oauth_account(
    auth: dmda.DomoAuth,
    account_id: str,
    share_payload: dict,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
    session: httpx.AsyncClient = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/data/v2/accounts/templates/share/{account_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=share_payload,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        debug_api=debug_api,
        session=session,
    )

    # does not appear to throw 500 error if user already has access
    # if res.status == 500 and res.response == "Internal Server Error":
    #     raise Account_AlreadyShared_Error(
    #         account_id=account_id,
    #         messagae=f"{res.response} - User may already have access to account",
    #         res=res,
    #     )

    if not res.is_success:
        raise Account_Share_Error(
            account_id=account_id,
            res=res,
        )

    return res

# %% ../../nbs/routes/account.ipynb 43
# v1 may have been deprecated.  used to be tied to group beta
@gd.route_function
async def share_account_v1(
    auth: dmda.DomoAuth,
    account_id: str,
    share_payload: dict,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
):
    """V1 API allows sharing with users ONLY.  it does not support sharing with groups and has a more limited set of share rights (owner or read) see ShareAccount_V1_AccessLevel vs v2"""
    
    url = f"https://{auth.domo_instance}.domo.com/api/data/v1/accounts/{account_id}/share"
    
    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=share_payload,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        debug_api=debug_api,
        session=session,
    )

    if res.status == 500 and res.response == "Internal Server Error":
        raise Account_AlreadyShared_Error(
            account_id=account_id,
            messagae=f"{res.response} - User may already have access to account",
            res=res,
        )

    if not res.is_success:
        raise Account_Share_Error(
            account_id=account_id,
            res=res,
        )

    return res

ShareAccount_V1_AccessLevel.share = staticmethod(share_account_v1)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/sandbox.ipynb.

# %% auto 0
__all__ = ['Sandbox_GET_Error', 'get_is_allow_same_instance_promotion_enabled', 'Sandbox_ToggleSameInstancePromotion_Error',
           'toggle_allow_same_instance_promotion', 'get_shared_repos', 'get_repo_from_id']

# %% ../../nbs/routes/sandbox.ipynb 2
import httpx
from typing import List
import pandas as pd


import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde

# %% ../../nbs/routes/sandbox.ipynb 5
class Sandbox_GET_Error(dmde.RouteError):
    def __init(self, res : rgd.ResponseGetData, message : str = None):
        super().__init__(res = res, message = message)

@gd.route_function
async def get_is_allow_same_instance_promotion_enabled(
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    debug_num_stacks_to_drop: int = 1,
    debug_api: bool = False,
    parent_class: str = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/version/v1/settings"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if return_raw:
        return res
    
    if not res.is_success:
        raise Sandbox_GET_Error(res = res)

    res.response = {
        "name": "allow_same_instance_promotion",
        "is_enabled": res.response["allowSelfPromotion"],
    }

    return res

# %% ../../nbs/routes/sandbox.ipynb 7
class Sandbox_ToggleSameInstancePromotion_Error(dmde.RouteError):
    def __init__(self, res: rgd.ResponseGetData, message: str = None, **kwargs):
        super().__init__(
            res=res,
            message=message,
            **kwargs
        )


@gd.route_function
async def toggle_allow_same_instance_promotion(
    is_enabled: bool,
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 1,
    debug_api: bool = False,
    parent_class: str = None,
):
    url = f"https://{auth.domo_instance}.domo.com/api/version/v1/settings"

    body = {"allowSelfPromotion": is_enabled}

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise Sandbox_ToggleSameInstancePromotion_Error(res=res)

    return res

# %% ../../nbs/routes/sandbox.ipynb 9
@gd.route_function
async def get_shared_repos(
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
    parent_class: str = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: bool = False,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/version/v1/repositories/search"

    body = {
        "query": {
            "offset": 0,
            "limit": 50,
            "fieldSearchMap": {},
            "sort": "lastCommit",
            "order": "desc",
            "filters": {"userId": None},
            "dateFilters": {},
        },
        "shared": False,
    }

    def arr_fn( res : rgd.ResponseGetData) -> List[dict]:
        return res.response['repositories'] 
    
    offset_params = {"offset": "offset", "limit": "limit"}

    def body_fn(skip, limit, body):
        body['query'].update({"offset" : skip, "limit": limit})
        return body

    res = await gd.looper(
        auth=auth,
        method="POST",
        url=url,
        arr_fn=arr_fn,
        body_fn = body_fn,
        body = body,
        loop_until_end=True,
        offset_params=offset_params,
        offset_params_in_body= True,
        session=session,
        return_raw = return_raw,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if return_raw:
        return res


    if not res.is_success:
        raise Sandbox_GET_Error(res = res)


    return res

# %% ../../nbs/routes/sandbox.ipynb 12
@gd.route_function
async def get_repo_from_id(
    auth: dmda.DomoAuth,
    repository_id: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
    session: httpx.AsyncClient = None,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/version/v1/repositories/{repository_id}"

    res =  await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        parent_class=parent_class,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )
    
    if not res.is_success:
        raise Sandbox_GET_Error(res = res)


    return res
    



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/client/95_DomoAuth.ipynb.

# %% auto 0
__all__ = ['DomoAuth', 'test_is_full_auth', 'DomoTokenAuth', 'DomoDeveloperAuth', 'DomoJupyterAuth', 'DomoJupyterFullAuth',
           'DomoJupyterTokenAuth', 'test_is_jupyter_auth']

# %% ../../nbs/client/95_DomoAuth.ipynb 3
from domolibrary.routes.auth import (
    AuthError,
    AccountLockedError,
    InvalidAuthTypeError,
    InvalidCredentialsError,
    InvalidInstanceError,
    NoAccessTokenReturned,
)

# %% ../../nbs/client/95_DomoAuth.ipynb 4
from dataclasses import dataclass, field
from typing import Optional, Union
from urllib.parse import urlparse

import domolibrary.client.DomoError as dmde
from abc import abstractmethod
import httpx

import domolibrary.client.Logger as lg

import domolibrary.routes.auth as auth_routes

# %% ../../nbs/client/95_DomoAuth.ipynb 8
@dataclass
class DomoAuth:
    """abstract DomoAuth class"""

    domo_instance: str
    url_manual_login: str

    auth_header: dict = field(repr=False)
    token_name: str
    token: str = field(repr=False)

    user_id: str
    is_valid_token: bool

    def __post_init__(self):
        """Initialize url_manual_login after instance creation."""
        self.url_manual_login = self._generate_manual_login(
            domo_instance=self.domo_instance
        )

        self.token_name = self.token_name or self.domo_instance

    @staticmethod
    def _generate_manual_login(domo_instance):
        return f"https://{domo_instance}.domo.com/auth/index?domoManualLogin=true"

    @classmethod
    def from_domo_instance(cls, domo_instance, token_name=None):
        """Create an instance using the provided domo_instance and optional token_name."""
        return cls(
            domo_instance=domo_instance,
            url_manual_login=cls._generate_manual_login(domo_instance),
            auth_header={},
            token_name=token_name,
            token=None,
            user_id=None,
            is_valid_token=False,
        )

    async def who_am_i(
        self,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ):
        """Perform an API call to identify the user associated with the token."""

        # if not self.token:
        #     await self.get_auth_token(debug_api=debug_api, session=session)

        # # if not self.auth_header:
        # #     self.generate_auth_header()

        res = await auth_routes.who_am_i(
            auth=self,
            parent_class=self.__class__.__name__,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        self.user_id = res.response["id"]

        return res

    async def elevate_otp(
        self,
        one_time_password: str,
        debug_api: bool = False,
        session: httpx.AsyncClient = None,
        debug_num_stacks_to_drop=2,
    ):
        """Elevate the authentication to include OTP (One-Time Password) if required."""

        res = await auth_routes.elevate_user_otp(
            auth=self,
            debug_api=debug_api,
            session=session,
            one_time_password=one_time_password,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        return res

    @abstractmethod
    async def get_auth_token(self) -> Union[str, None]:
        """placeholder method"""
        pass

    @abstractmethod
    def _generate_auth_header(self) -> Union[dict, None]:
        """returns auth header appropriate for this authentication method"""
        pass

    def generate_auth_header(self) -> Union[dict, None]:
        """Generate the authentication header using the specialized method."""
        return self._generate_auth_header()

    async def print_is_token(
        self,
        debug_api: bool = False,
        token_name=None,
        session: httpx.AsyncClient = None,
    ) -> bool:
        """Print token status and return True if token is valid, otherwise False."""
        self.token_name = token_name or self.token_name

        if not self.token:
            await self.get_auth_token(debug_api=debug_api, session=session)

        token_str = f"{self.token_name} " or ""
        if not self.token:
            print(f"🚧 failed to retrieve {token_str}token from {self.domo_instance}")
            return False

        print(f"🎉 {token_str}token retrieved from {self.domo_instance} ⚙️")
        return True

# %% ../../nbs/client/95_DomoAuth.ipynb 12
@dataclass
class DomoFullAuth(DomoAuth):
    """mix requied parameters for DomoFullAuth"""

    domo_username: str
    domo_password: str = field(repr=False)

    def __init__(
        self, domo_instance, domo_username, domo_password, token_name: str = None
    ):
        """Initialize DomoFullAuth using username and password credentials."""
        self.domo_username = domo_username
        self.domo_password = domo_password

        super().__init__(
            domo_instance=domo_instance,
            url_manual_login=super()._generate_manual_login(domo_instance),
            auth_header={},
            token_name=token_name,
            token=None,
            user_id=None,
            is_valid_token=False,
        )

    def _generate_auth_header(self) -> dict:
        """Generate the full authentication header specific to product APIs."""
        self.auth_header = {"x-domo-authentication": self.token}
        return self.auth_header

    def generate_auth_header(self) -> dict:
        """Return the generated full authentication header."""
        return self._generate_auth_header()

    async def get_auth_token(
        self,
        session: Optional[httpx.AsyncClient] = None,
        return_raw: bool = False,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ) -> str:
        """Retrieve the authentication token from product APIs using the provided credentials."""

        res = await auth_routes.get_full_auth(
            auth=None,
            domo_instance=self.domo_instance,
            domo_username=self.domo_username,
            domo_password=self.domo_password,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        )

        if return_raw:
            return res

        self.is_valid_token = True
        self.token = str(res.response.get("sessionToken"))

        self.token_name = self.token_name or "full_auth"

        self.generate_auth_header()

        return self.token

# %% ../../nbs/client/95_DomoAuth.ipynb 17
def test_is_full_auth(
    auth, function_name=None, num_stacks_to_drop=1  # pass q for route pass 2 for class
):
    """Test that the provided object is a DomoFullAuth instance."""
    tb = lg.get_traceback(num_stacks_to_drop=num_stacks_to_drop)

    function_name = function_name or tb.function_name

    if auth.__class__.__name__ != "DomoFullAuth":
        raise InvalidAuthTypeError(
            function_name=function_name,
            domo_instance=auth.domo_instance,
            required_auth_type=DomoFullAuth,
        )

# %% ../../nbs/client/95_DomoAuth.ipynb 19
@dataclass
class DomoTokenAuth(DomoAuth):
    domo_access_token: str = field(repr=False)

    """
    use for access_token authentication.
    Tokens are generated in domo > admin > access token
    Necessary in cases where direct sign on is not permitted
    """

    def __init__(
        self, domo_access_token: str, domo_instance: str, token_name: str = None
    ):
        """Initialize DomoTokenAuth with the provided access token for product APIs."""
        self.domo_access_token = domo_access_token

        super().__init__(
            domo_instance=domo_instance,
            url_manual_login=super()._generate_manual_login(domo_instance),
            auth_header={},
            token_name=token_name,
            token=domo_access_token,
            user_id=None,
            is_valid_token=False,
        )

    def _generate_auth_header(self) -> dict:
        """Generate the authentication header for access token based authentication."""
        self.auth_header = {
            "x-domo-developer-token": self.token or self.domo_access_token
        }

        return self.auth_header

    def generate_auth_header(self) -> dict:
        """Return the generated access token authentication header."""
        return self._generate_auth_header()

    async def get_auth_token(
        self,
        session: Optional[httpx.AsyncClient] = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ) -> str:
        """Retrieve the access token, updating internal attributes as necessary."""

        if not self.user_id:
            await self.who_am_i(
                session=session,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
            )

        self.is_valid_token = True
        self.token = self.domo_access_token
        self.token_name = self.token_name or "token_auth"

        return self.token

# %% ../../nbs/client/95_DomoAuth.ipynb 25
@dataclass
class DomoDeveloperAuth(DomoAuth):
    domo_client_id: str
    domo_client_secret: str = field(repr=False)

    def __init__(
        self,
        domo_client_id,
        domo_client_secret,
        token_name: str = None,
        domo_instance: str = None,
    ):
        """Initialize DomoDeveloperAuth with client credentials for platform APIs."""
        self.domo_client_id = domo_client_id
        self.domo_client_secret = domo_client_secret

        super().__init__(
            domo_instance=domo_instance,
            url_manual_login=super()._generate_manual_login(domo_instance),
            auth_header={},
            token_name=token_name,
            token=None,
            user_id=None,
            is_valid_token=False,
        )

    def _generate_auth_header(self) -> dict:
        """Generate the authentication header for developer token authentication."""
        self.auth_header = {"Authorization": "bearer " + self.token}

        return self.auth_header

    def generate_auth_header(self) -> dict:
        """Return the generated developer token authentication header."""
        return self._generate_auth_header()

    async def get_auth_token(
        self,
        session: Optional[httpx.AsyncClient] = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ) -> str:
        """Retrieve the developer token using client credentials and update internal attributes."""

        res = await auth_routes.get_developer_auth(
            auth=None,
            domo_client_id=self.domo_client_id,
            domo_client_secret=self.domo_client_secret,
            session=session,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        self.is_valid_token = True

        self.token = str(res.response.get("access_token"))
        self.user_id = res.response.get("userId")
        self.domo_instance = res.response.get("domain")

        self._generate_manual_login()

        self.generate_auth_header()

        self.token_name = self.token_name or "developer_auth"

        return self.token

# %% ../../nbs/client/95_DomoAuth.ipynb 29
@dataclass
class _DomoJupyter_Optional:
    def __post_init__(self):

        self.jupyter_token = self.jupyter_token or input(
            "jupyter token: # retrieve this by monitoring domo jupyter network traffic.  it is the Authorization header"
        )
        self.service_location = self.service_location or input(
            "service_location:  # retrieve from domo jupyter env"
        )
        self.service_prefix = self.service_prefix or input(
            "service prefix: # retrieve from domo jupyter env"
        )

        self._test_prereq()
        self._generate_manual_login(domo_instance=self.domo_instance)
        
        self.generate_auth_header()


@dataclass
class _DomoJupyter_Required:
    jupyter_token: str
    service_location: str
    service_prefix: str

    def get_jupyter_token_flow(self):
        """stub"""
        print("hello world i am a jupyter_token")

    def _test_prereq(self):
        if not self.jupyter_token:
            raise Exception("DomoJupyterAuth objects must have a jupyter_token")

        if not self.service_location:
            raise Exception("DomoJupyterAuth objects must have a service_location")

        if not self.service_prefix:
            raise Exception("DomoJupyterAuth objects must have a service_prefix")

        if (
            not self.jupyter_token
            or not self.service_location
            or not self.service_prefix
        ):
            raise Exception(
                "DomoJupyterAuth objects must have jupyter_token, service_location and service_prefix"
            )

# %% ../../nbs/client/95_DomoAuth.ipynb 30
@dataclass
class DomoJupyterAuth(_DomoJupyter_Optional, _DomoJupyter_Required):
    """base class"""

# %% ../../nbs/client/95_DomoAuth.ipynb 32
@dataclass
class DomoJupyterFullAuth(_DomoJupyter_Optional, DomoFullAuth, _DomoJupyter_Required):

    @classmethod
    def convert_auth(
        cls, auth: DomoFullAuth, jupyter_token, service_location, service_prefix
    ):
        """converts DomoFullAuth to DomoJupyterFullAuth
        i.e. adds DomoJupyter specific auth fields
        eventually can add DomoJupyter specific auth flow for generating auth token
        """
        return cls(
            domo_instance=auth.domo_instance,
            domo_username=auth.domo_username,
            domo_password=auth.domo_password,
            jupyter_token=jupyter_token,
            service_location=service_location,
            service_prefix=service_prefix,
            url_manual_login=auth.url_manual_login,
            auth_header=auth.auth_header,
            token_name=auth.token_name,
            token=auth.token,
            user_id=auth.user_id,
            is_valid_token=auth.is_valid_token,
        )

    def generate_auth_header(self) -> dict:

        self._generate_auth_header()

        self.auth_header.update(
            {
                "authorization": f"Token {self.jupyter_token}",
            }
        )

        return self.auth_header

# %% ../../nbs/client/95_DomoAuth.ipynb 36
@dataclass
class DomoJupyterTokenAuth(_DomoJupyter_Optional, DomoTokenAuth, _DomoJupyter_Required):
    @classmethod
    def convert_auth(
        cls, auth: DomoTokenAuth, jupyter_token, service_location, service_prefix
    ):
        """converts DomoTokenAuth to DomoJupyterTokenAuth
        i.e. adds DomoJupyter specific auth fields
        eventually can add DomoJupyter specific auth flow for generating auth token
        """
        return cls(
            domo_instance=auth.domo_instance,
            domo_access_token=auth.domo_access_token,
            jupyter_token=jupyter_token,
            service_location=service_location,
            service_prefix=service_prefix,
            url_manual_login=auth.url_manual_login,
            auth_header=auth.auth_header,
            token_name=auth.token_name,
            token=auth.token,
            user_id=auth.user_id,
            is_valid_token=auth.is_valid_token,
        )

    def generate_auth_header(self) -> dict:
        self._generate_auth_header()

        self.auth_header.update(
            {
                "authorization": f"Token {self.jupyter_token}",
            }
        )

        return self.auth_header

# %% ../../nbs/client/95_DomoAuth.ipynb 40
def test_is_jupyter_auth(
    auth: DomoJupyterAuth,
    function_name=None,
    required_auth_type_ls=[DomoJupyterFullAuth, DomoJupyterTokenAuth],
):
    """Test that the provided object is a valid Jupyter authentication instance."""
    tb = lg.get_traceback()

    if auth.__class__.__name__ not in [
        auth_type.__name__ for auth_type in required_auth_type_ls
    ]:
        raise InvalidAuthTypeError(
            function_name=tb.function_name,
            domo_instance=auth.domo_instance,
            required_auth_type_ls=required_auth_type_ls,
        )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/publish.ipynb.

# %% auto 0
__all__ = ['GET_Publish_Error', 'CRUD_Publish_Error', 'search_publications', 'get_publication_by_id', 'generate_publish_body',
           'create_publish_job', 'udpate_publish_job', 'get_publish_subscriptions', 'get_subscription_summaries',
           'get_subscriber_content_details', 'get_subscription_invititations', 'get_subscriber_domains',
           'add_subscriber_domain', 'accept_invite_by_id', 'accept_invite_by_id_v2', 'refresh_publish_jobs']

# %% ../../nbs/routes/publish.ipynb 2
from typing import List
import httpx

import domolibrary.client.get_data as gd
import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as de

# %% ../../nbs/routes/publish.ipynb 6
class GET_Publish_Error(de.RouteError):
    def __init__(self, res: rgd.ResponseGetData, message:str = None):
        super().__init__(
            res = res,
            message = message
        )


class CRUD_Publish_Error(de.DomoError):
    def __init__(self, res: rgd.ResponseGetData, message : str = None):
        super().__init__(
            res = res
        )

# %% ../../nbs/routes/publish.ipynb 7
@gd.route_function
async def search_publications(
    auth: dmda.DomoAuth,
    search_term: str = None,
    limit=100,
    offset=0,
    session: httpx.AsyncClient = None,
    debug_loop :bool= False,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    return_raw: bool = False
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/publication/summaries"

    offset_params = {"limit": "limit", "offset": "offset"}

    params = {}
    if search_term:
        params.update( {"searchTerm": search_term})
    
    def arr_fn(res: rgd.ResponseGetData):
        return  res.response

    res = await gd.looper(
        auth=auth,
        method="GET",
        limit = limit,
        skip = offset,
        arr_fn = arr_fn,
        fixed_params = params,
        offset_params=offset_params,
        loop_until_end=True,
        debug_loop=debug_loop,
        url=url,
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        return_raw= return_raw
    )

    if return_raw: 
        return res

    if not res.is_success:
        raise GET_Publish_Error(res)

    return res

# %% ../../nbs/routes/publish.ipynb 9
@gd.route_function
async def get_publication_by_id(
    auth: dmda.DomoAuth,
    publication_id: str,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
    timeout=10,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/publication/{publication_id}"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        session=session,
        timeout=timeout,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise GET_Publish_Error(res)

    return res


# generate publish body

# %% ../../nbs/routes/publish.ipynb 11
def generate_publish_body(
    url: str,
    sub_domain_ls: List[str],
    content_ls: List[str],
    name: str,
    description: str,
    unique_id: str,
    is_new: bool,
):
    if not sub_domain_ls:
        sub_domain_ls = []

    if not content_ls:
        content_ls = []

    body = {
        "id": unique_id,
        "name": name,
        "description": description,
        "domain": url,
        "content": content_ls,
        "subscriberDomain": sub_domain_ls,
        "new": str(is_new).lower(),
    }

    return body

# %% ../../nbs/routes/publish.ipynb 13
# Creating publish job for a specific subscriber
@gd.route_function
async def create_publish_job(
    auth: dmda.DomoAuth,
    body: dict,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:
    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/publication"

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise CRUD_Publish_Error(res)

    return res

# %% ../../nbs/routes/publish.ipynb 14
# Updating existing publish job with content
@gd.route_function
async def udpate_publish_job(
    auth: dmda.DomoAuth,
    publication_id: str,
    body: dict,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
) -> rgd.ResponseGetData:

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/publication/{publication_id}"

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise CRUD_Publish_Error(res)
    return res


# %% ../../nbs/routes/publish.ipynb 16
@gd.route_function
async def get_publish_subscriptions(
    auth: dmda.DomoAuth,
    publish_id: str,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
) -> rgd.ResponseGetData:
    """retrieves a summary of existing subscriptions"""
    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/publications/summaries/{publish_id}/subscriptions"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise GET_Publish_Error(res)

    return res

# %% ../../nbs/routes/publish.ipynb 18
@gd.route_function
async def get_subscription_summaries(
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
) -> rgd.ResponseGetData:
    """retrieves a summary of existing subscriptions"""

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/subscription/summaries"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise GET_Publish_Error(res)
    return res

# %% ../../nbs/routes/publish.ipynb 22
@gd.route_function
async def get_subscriber_content_details(
    auth: dmda.DomoAuth,
    publication_id,
    subscriber_instance: str,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
):
    if not subscriber_instance.endswith(".domo.com"):
        subscriber_instance = f"{subscriber_instance}.domo.com"

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/associations/{publication_id}/subscribers/{subscriber_instance}"

    res = await gd.get_data(
        method="get",
        url=url,
        auth=auth,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise GET_Publish_Error(res)

    return res

# %% ../../nbs/routes/publish.ipynb 24
@gd.route_function
async def get_subscription_invititations(
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:
    """retrieves a list of subscription invitations"""

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/subscription/invites"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise GET_Publish_Error(res)
    return res

# %% ../../nbs/routes/publish.ipynb 25
@gd.route_function
async def get_subscriber_domains(
    auth: dmda.DomoAuth,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:
    """retrieves a list of subsriber domains"""

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/proxy_user?parts=SUBSCRIPTION_COUNT"

    res = await gd.get_data(
        auth=auth,
        method="GET",
        url=url,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise GET_Publish_Error(res)
    return res

@gd.route_function
async def add_subscriber_domain(
    auth: dmda.DomoAuth,
    domain: str,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:
    """adds subscriber domain to the list"""
    if ".domo.com" not in domain: 
        domain = domain+".domo.com"
    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/proxy_user/domain/{domain}/"
    body = {"tenantId":domain}
    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    if not res.is_success:
        raise GET_Publish_Error(res)
    return res

# %% ../../nbs/routes/publish.ipynb 28
@gd.route_function
async def accept_invite_by_id(
    auth: dmda.DomoAuth,
    subscription_id: str,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    parent_class: str = None,
) -> rgd.ResponseGetData:
    """this takes get_subscription_invites_list into account and accepts - not instant"""

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/subscription/{subscription_id}"

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise CRUD_Publish_Error(res)
    return res

# %% ../../nbs/routes/publish.ipynb 29
@gd.route_function
async def accept_invite_by_id_v2(
    auth: dmda.DomoAuth,
    publication_id: str,
    owner_id: str,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
) -> rgd.ResponseGetData:
    """this takes get_subscription_invites_list into account and accepts - not instant"""

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/subscriptions/v2"

    body = {
        "publicationId": publication_id,
        "customerId": "",
        "domain": "",
        "groupIds": [],
        "userId": owner_id,
        "userIds": [],
    }

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise CRUD_Publish_Error(res)

    return res

# %% ../../nbs/routes/publish.ipynb 30
@gd.route_function
async def refresh_publish_jobs(
    auth: dmda.DomoAuth,
    publish_ids: list,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 1,
    parent_class: str = None,
) -> rgd.ResponseGetData:
    """Refreshing list of publish jobs. Typically "instance" = publisher instance"""

    url = f"https://{auth.domo_instance}.domo.com/api/publish/v2/publication/refresh"

    body = {"publicationIds": publish_ids}

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise CRUD_Publish_Error(res)

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoUser.ipynb.

# %% auto 0
__all__ = ['CreateUser_MissingRole', 'DownloadAvatar_NoAvatarKey', 'DomoUser', 'DomoUsers', 'DomoUser_NoSearch']

# %% ../../nbs/classes/50_DomoUser.ipynb 2
from domolibrary.routes.user import (
    UserProperty,
    UserProperty_Type,
    GetUser_Error,
    SearchUser_NoResults,
    User_CrudError,
    ResetPassword_PasswordUsed,
    DeleteUser_Error,
    DownloadAvatar_Error,
    UserProperty_Type
)


from ..routes.instance_config_sso import (SSO_AddUserDirectSignonError)
from ..utils.Image import Image, are_same_image, domo_default_img

# %% ../../nbs/classes/50_DomoUser.ipynb 3
import datetime as dt
from dataclasses import dataclass, field
from typing import Optional, List, Any, Union
import httpx

from nbdev.showdoc import patch_to
import asyncio

import domolibrary.utils.DictDot as util_dd
from domolibrary.utils.convert import (test_valid_email, convert_epoch_millisecond_to_datetime)

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.Logger as lc
import domolibrary.client.DomoError as dmde
import domolibrary.routes.user as user_routes
import domolibrary.routes.instance_config_sso as sso_routes

# %% ../../nbs/classes/50_DomoUser.ipynb 7
class CreateUser_MissingRole(dmde.DomoError):
    def __init__(self, domo_instance, email_address):
        super().__init__(
            domo_instance=domo_instance,
            message=f"error creating user {email_address} missing role_id",
        )


class DownloadAvatar_NoAvatarKey(dmde.DomoError):
    def __init__(
        self,
        domo_instance,
        user_id,
    ):
        super().__init__(
            domo_instance,
            message=f"This profile {user_id} doesn't have an avatar uploaded - unable to download",
        )

# %% ../../nbs/classes/50_DomoUser.ipynb 9
@dataclass
class DomoUser:
    """a class for interacting with a Domo User"""

    auth: dmda.DomoAuth = field(repr=False)

    id: str
    display_name: str = None
    email_address: str = None
    role_id: str = None
    department: str = None
    title: str = None
    avatar_key: str = None
    password : str = field(repr = False, default = None)

    phone_number: str = None
    web_landing_page: str = None
    web_mobile_landing_page: str = None
    employee_id: str = None
    employee_number: str = None
    hire_date: str = None
    reports_to: str = None

    publisher_domain: str = None
    subscriber_domain: str = None
    virtual_user_id: str = None

    created_dt : dt.datetime = None
    last_activity_dt: dt.datetime = None

    custom_attributes: dict = field(default_factory=dict)

    role: Any = None  # DomoRole
    domo_api_clients : List[Any] = None
    domo_access_tokens : List[Any] = None


    def __post_init__(self):
        self.id = str(self.id)

    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False

        return self.id == other.id

    @classmethod
    def _from_search_json(cls, auth, user_obj):
        user_dd = util_dd.DictDot(user_obj)

        return cls(
            auth=auth,
            id=str(user_dd.id or user_dd.userId),
            display_name=user_dd.displayName,
            title=user_dd.title,
            department=user_dd.department,
            email_address=user_dd.emailAddress or user_dd.email,
            role_id=user_dd.roleId,
            avatar_key=user_dd.avatarKey,
            phone_number=user_dd.phoneNumber,
            web_landing_page=user_dd.webLandingPage,
            web_mobile_landing_page=user_dd.webMobileLandingPage,
            employee_id=user_dd.employeeId,
            employee_number=user_dd.employeeNumber,
            hire_date=user_dd.hireDate,
            reports_to=user_dd.reportsTo,
            created_dt = convert_epoch_millisecond_to_datetime(user_dd.created),
            last_activity_dt = convert_epoch_millisecond_to_datetime(user_dd.lastActivity)
        )

    @classmethod
    def _from_virtual_json(cls, auth, user_obj):
        user_dd = util_dd.DictDot(user_obj)

        return cls(
            id=user_dd.id,
            auth=auth,
            publisher_domain=user_dd.publisherDomain,
            subscriber_domain=user_dd.subscriberDomain,
            virtual_user_id=user_dd.virtualUserId,
        )

    @classmethod
    def _from_bootstrap_json(cls, auth, user_obj):
        dd = user_obj
        if isinstance(user_obj, dict):
            dd = util_dd.DictDot(user_obj)

        return cls(id=dd.id, display_name=dd.displayName, auth=auth)

# %% ../../nbs/classes/50_DomoUser.ipynb 11
@patch_to(DomoUser)
async def get_role(
    self: DomoUser,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
):
    import domolibrary.classes.DomoRole as dmr

    self.role = await dmr.DomoRole.get_by_id(
        role_id=self.role_id,
        auth=self.auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )

    return self.role

# %% ../../nbs/classes/50_DomoUser.ipynb 12
@patch_to(DomoUser, cls_method=True)
async def get_by_id(
    cls: DomoUser,
    user_id,
    auth: dmda.DomoAuth,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
):
    """
    searches and returns a domo user
    will throw an error if no user returned with an option to suppress_no_results_error
    """

    res = await user_routes.get_by_id(
        auth=auth,
        user_id=user_id,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
        parent_class=cls.__name__,
    )

    if return_raw:
        return res

    if not res.is_success:
        return None

    domo_user = cls._from_search_json(user_obj=res.response, auth=auth)
    
    try:
        await domo_user.get_role(debug_api = debug_api, debug_num_stacks_to_drop = debug_num_stacks_to_drop, session = session)
    
    except Exception as e:
        print(e)


    return domo_user

# %% ../../nbs/classes/50_DomoUser.ipynb 15
@patch_to(DomoUser)
async def download_avatar(
    self: DomoUser,
    pixels: int = 300,
    folder_path="./images",
    img_name=None,  # will default to user_id
    auth: dmda.DomoAuth = None,
    is_download_image: bool = True,  # option to prevent downloading the image file
    debug_api: bool = False,
    return_raw: bool = False,
):
    """downloads a user's avatar to a folder
    and returns the byte representation of the image
    """
    auth = auth or self.auth

    # if not self.avatar_key:
    #     raise DownloadAvatar_NoAvatarKey(
    #         domo_instance=auth.domo_instance, user_id=self.id
    #     )

    res = await user_routes.download_avatar(
        auth=self.auth,
        user_id=self.id,
        pixels=pixels,
        folder_path=folder_path,
        img_name=img_name,
        is_download_image=is_download_image,
        debug_api=debug_api,
    )

    if return_raw:
        return res

    self.avatar = Image.from_bytestr(data=res.response)

    return self.avatar

    # return res.response

# %% ../../nbs/classes/50_DomoUser.ipynb 19
@patch_to(DomoUser)
async def update_properties(
    self: DomoUser,
    property_ls: List[
        UserProperty
    ],  # use the UserProperty class to define a list of user properties to update, see user route documentation to see a list of UserProperty_Types that can be updated
    return_raw: bool = False,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    auth = auth or self.auth

    res = await user_routes.update_user(
        auth=auth,
        user_id=self.id,
        user_property_ls=property_ls,
        debug_api=debug_api,
        session=session,
    )
    if return_raw:
        return res

    self = await DomoUser.get_by_id(user_id=self.id, auth=auth)

    return self


@patch_to(DomoUser)
async def set_user_landing_page(
    self: DomoUser,
    page_id: str,
    user_id: str = None,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
):
    res = await user_routes.set_user_landing_page(
        auth=auth or self.auth,
        page_id=page_id,
        user_id=self.id or user_id,
        debug_api=debug_api,
    )

    return res

# %% ../../nbs/classes/50_DomoUser.ipynb 22
@patch_to(DomoUser, cls_method= True)
async def create(
    cls,
    auth : dmda.DomoAuth,
    display_name,
    email_address,
    role_id,
    password: str = None,
    send_password_reset_email: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    session: httpx.AsyncClient = None,
):
    """class method that creates a new Domo user"""

    res = await user_routes.create_user(
        auth=auth,
        display_name=display_name,
        email_address=email_address,
        role_id=role_id,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    domo_user = await DomoUser.get_by_id(
        auth=auth,
        user_id=res.response.get("id") or res.response.get("userId"),
    )

    if password:
        await domo_user.reset_password(new_password=password)

    elif send_password_reset_email:
        await domo_user.request_password_reset(
            domo_instance=auth.domo_instance, email=domo_user.email_address
        )

    return domo_user



@patch_to(DomoUser)
async def delete(
    self: DomoUser,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=2,
    parent_class=None,
):

    res = await user_routes.delete_user(
        auth=self.auth,
        user_id=self.id,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
    )

    return res

# %% ../../nbs/classes/50_DomoUser.ipynb 26
@patch_to(DomoUser)
async def reset_password(
    self: DomoUser,
    new_password: str,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop: int = 2,
):
    """reset your password, will respect password restrictions set up in the Domo UI"""

    res = await user_routes.reset_password(
        auth=self.auth,
        user_id=self.id,
        new_password=new_password,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class = self.__class__.__name__
    )

    self.password = new_password

    return res


@patch_to(DomoUser, cls_method=True)
async def request_password_reset(
    cls,
    domo_instance: str,
    email: str,
    locale: str = "en-us",
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop : int = 2
):
    """request password reset email.  Note: does not require authentication."""

    return await user_routes.request_password_reset(
        domo_instance=domo_instance,
        email=email,
        locale=locale,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop = debug_num_stacks_to_drop,
        parent_class = cls.__name__
    )

# %% ../../nbs/classes/50_DomoUser.ipynb 27
@patch_to(DomoUser)
async def upload_avatar(
    self,
    avatar: Image,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):
    avatar.crop_square()

    res = await user_routes.upload_avatar(
        auth=self.auth,
        user_id=self.id,
        img_bytestr=avatar.to_bytes(),
        img_type=avatar.format,
        debug_api=debug_api,
        parent_class=self.__class__.__name__,
        session=session,
    )

    if return_raw:
        return res

    await asyncio.sleep(2)

    return await self.download_avatar(debug_api=debug_api)

# %% ../../nbs/classes/50_DomoUser.ipynb 29
@patch_to(DomoUser)
async def upsert_avatar(
    self,
    avatar: Image,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):
    avatar.crop_square()

    res = "images are the same"
    if not are_same_image(domo_default_img, avatar):

        res = await user_routes.upload_avatar(
            auth=self.auth,
            user_id=self.id,
            img_bytestr=avatar.to_bytes(),
            img_type=avatar.format,
            debug_api=debug_api,
            parent_class=self.__class__.__name__,
            session=session,
        )

        if return_raw:
            return res

    await asyncio.sleep(2)

    return await self.download_avatar(debug_api=debug_api)

# %% ../../nbs/classes/50_DomoUser.ipynb 31
@patch_to(DomoUser)
async def toggle_direct_signon_access(self : DomoUser,
                                      is_enable_direct_signon : bool = True,
                                      session: httpx.AsyncClient = None,
                                      debug_api : bool =False,
                                      debug_num_stacks_to_drop : int= 2 ):
    
    res = await sso_routes.toggle_user_direct_signon_access(auth= self.auth,
        user_id_ls = [self.id],
        is_enable_direct_signon =is_enable_direct_signon,
        session = session,
        debug_api = debug_api,
        parent_class= self.__class__.__name__,
        debug_num_stacks_to_drop= debug_num_stacks_to_drop)
    
    return res

# %% ../../nbs/classes/50_DomoUser.ipynb 32
@patch_to(DomoUser)
async def get_api_clients(
    self,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
    debug_num_stacks_to_drop=2,
    return_raw: bool = False,
):
    """
    retrieves Client_IDs for this user (assuming the authenticated user has manage rights).
    Note : the values will be masked, raw text values can only be retrieved via the UI
    """

    import domolibrary.classes.DomoInstanceConfig_ApiClient as dicli

    api_clients = dicli.ApiClients(auth=auth)
    domo_clients = await api_clients.get(
        session=session,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop+1,
    )

    if return_raw: return res

    domo_clients = [
        domo_client
        for domo_client in domo_clients
        if domo_client.domo_user.id == self.id
    ]

    if not domo_clients:
        print(f"Domo User {self.id} - {self.display_name} does not have any Client_IDs")
        return False

    self.domo_api_clients = domo_clients

    return self.domo_api_clients

# %% ../../nbs/classes/50_DomoUser.ipynb 33
@patch_to(DomoUser)
async def get_access_tokens(
    self,
    debug_api: bool = False,
    debug_num_stacks_to_drop : int = 2,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):
    import domolibrary.classes.DomoAccessToken as dmat

    domo_config = dmat.DomoAccessTokens(auth = self.auth)
    domo_tokens = await domo_config.get(
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop= debug_num_stacks_to_drop + 1
    )

    if return_raw:
        return domo_tokens
    
    domo_tokens = [domo_token for domo_token in domo_tokens if domo_token.owner.id == self.id]

    if not domo_tokens:
        print(f"Domo User {self.id} - {self.display_name} does not have any access tokens")
        return []
     
    self.domo_access_tokens = domo_tokens

    return self.domo_access_tokens

# %% ../../nbs/classes/50_DomoUser.ipynb 36
@dataclass
class DomoUsers:
    """a class for searching for Users"""

    auth: dmda.DomoAuth = field(repr=False)

    users: List[DomoUser] = None
    logger: Optional[lc.Logger] = None

    @classmethod
    def _users_to_domo_user(cls, user_ls, auth: dmda.DomoAuth):
        return [
            DomoUser._from_search_json(auth=auth, user_obj=user_obj)
            for user_obj in user_ls
        ]

    @classmethod
    def _users_to_virtual_user(cls, user_ls, auth: dmda.DomoAuth):
        return [
            DomoUser._from_virtual_json(auth=auth, user_obj=user_obj)
            for user_obj in user_ls
        ]

    def _generate_logger(self, logger: Optional[lc.Logger] = None):
        self.logger = logger or self.logger or lc.Logger(app_name="domo_users")

    @staticmethod
    def _util_match_domo_users_to_emails(
        domo_users: list[DomoUser], user_email_ls: list[str]
    ) -> list:
        """pass in an array of user emails to match against an array of Domo User"""

        return [
            domo_user
            for domo_user in domo_users
            if domo_user.email_address.lower() in [email.lower() for email in user_email_ls]
        ]

    @staticmethod
    def _util_match_users_obj_to_emails(
        user_ls: list[dict], user_email_ls: list[str]
    ) -> list:
        """pass in an array of user emails to match against an array of Domo User"""

        return [
            obj
            for obj in user_ls
            if obj.get("emailAddress").lower() in [email.lower() for email in user_email_ls]
        ]


# %% ../../nbs/classes/50_DomoUser.ipynb 38
@patch_to(DomoUsers)
async def get(
    self,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session : httpx.AsyncClient = None,
) -> List[DomoUser]:
    """retrieves all users from Domo"""

    res = await user_routes.get_all_users(
        auth=self.auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
        session = session
    )

    if return_raw:
        return res

    self.users = self._users_to_domo_user(user_ls=res.response, auth=self.auth)
    return self.users

@patch_to(DomoUsers, cls_method=True)
async def all_users(
    cls: DomoUsers,
    auth: dmda.DomoAuth,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session : httpx.AsyncClient = None,
    logger: Optional[lc.Logger] = None,
) -> List[DomoUser]:
    """retrieves all users from Domo"""

    res = await user_routes.get_all_users(
        auth=auth,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=cls.__name__,
        session = session
    )

    if return_raw:
        return res

    if not res.is_success:
        return None

    users_ls = res.response

    return cls._users_to_domo_user(user_ls=users_ls, auth=auth)

# %% ../../nbs/classes/50_DomoUser.ipynb 40
class DomoUser_NoSearch(dmde.ClassError):
    def __init__(self, message : str, domo_instance , cls_instance = None, cls = None):
        super().__init__(cls= cls, cls_instance=cls_instance, message = message,
                         entity_id = domo_instance
                         )

# %% ../../nbs/classes/50_DomoUser.ipynb 41
@patch_to(DomoUsers)
async def search_by_email(
    self,
    email: Union[str, list],
    only_allow_one: bool = True,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    return_raw: bool = False,
    suppress_no_results_error: bool = False,
    session: httpx.AsyncClient = None,
) -> list:

    emails = [email] if isinstance(email, str) else email

    try:
        res = await user_routes.search_users_by_email(
            user_email_ls=emails,
            auth=self.auth,
            return_raw=return_raw,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )

    except SearchUser_NoResults as e:
        if suppress_no_results_error:
            return None

        raise e from e

    if return_raw:
        return res

    domo_users = self._users_to_domo_user(res.response, auth=self.auth)

    if not only_allow_one:
        return domo_users

    domo_users = self._util_match_domo_users_to_emails(domo_users, emails)

    if not domo_users:
        raise DomoUser_NoSearch(
            cls_instance=self, message=f'unable to find {",".join(emails)}',
            domo_instance = self.auth.domo_instance
        )

    return domo_users[0]


@patch_to(DomoUsers, cls_method=True)
async def by_email(
    cls: DomoUsers,
    email_ls: list[str],
    auth: dmda.DomoAuth,
    only_allow_one: bool = True,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    return_raw: bool = False,
    suppress_no_results_error: bool = False,
    session: httpx.AsyncClient = None,
) -> list:

    try:
        res = await user_routes.search_users_by_email(
            user_email_ls=email_ls,
            auth=auth,
            return_raw=False,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=cls.__name__,
            session=session,
        )
    except SearchUser_NoResults as e:
        if suppress_no_results_error:
            return None

        raise e

    if return_raw:
        return res

    domo_users = cls._users_to_domo_user(res.response, auth=auth)

    if not only_allow_one:
        return domo_users

    domo_users = cls._util_match_domo_users_to_emails(domo_users, email_ls)

    if not domo_users:
        raise DomoUser_NoSearch(cls=cls, message=f'unable to find {",".join(email_ls)}', domo_instance = auth.domo_instance)

    return domo_users[0]

@patch_to(DomoUsers, cls_method=True)
async def by_id(
    cls: DomoUsers,
    auth: dmda.DomoAuth,
    user_ids: list[str],  # can search for one or multiple users
    suppress_no_results_error: bool = False,
    only_allow_one: bool = True,
    debug_num_stacks_to_drop=2,
    debug_api: bool = False,
    return_raw: bool = False,
    session : httpx.AsyncClient = None
) -> list:
    
    res = None

    try:
        res = await user_routes.search_users_by_id(
            return_raw=False,
            user_ids=user_ids,
            debug_api=debug_api,
            auth=auth,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=cls.__name__,
            session = session
        )

    except SearchUser_NoResults as e:
        if suppress_no_results_error:
            return None

    if return_raw:
        return res

    domo_users = cls._users_to_domo_user(user_ls=res.response, auth=auth)

    if only_allow_one:
        return domo_users[0]

    return domo_users

# %% ../../nbs/classes/50_DomoUser.ipynb 44
@patch_to(DomoUsers, cls_method=True)
async def virtual_user_by_subscriber_instance(
    cls: DomoUsers,
    subscriber_instance_ls: str,
    auth: dmda.DomoAuth,
    debug_api: bool = False,
    return_raw: bool = False,
):
    res = await user_routes.search_virtual_user_by_subscriber_instance(
        auth=auth,
        subscriber_instance_ls=subscriber_instance_ls,
        debug_api=debug_api,
    )

    if return_raw:
        return res


    domo_users = cls._users_to_virtual_user(res.response, auth=auth)
    return domo_users

# %% ../../nbs/classes/50_DomoUser.ipynb 47
@patch_to(DomoUsers, cls_method=True)
async def create_user(
    cls: DomoUsers,
    auth: dmda.DomoAuth,
    display_name,
    email_address,
    role_id,
    password: str = None,
    send_password_reset_email: bool = False,
    debug_api: bool = False,
    session: httpx.AsyncClient = None,
):
    """class method that creates a new Domo user"""

    res = await user_routes.create_user(
        auth=auth,
        display_name=display_name,
        email_address=email_address,
        role_id=role_id,
        debug_api=debug_api,
        session=session,
    )

    domo_user = await DomoUser.get_by_id(
        auth=auth,
        user_id=res.response.get("id") or res.response.get("userId"),
    )

    if password:
        await domo_user.reset_password(new_password=password)

    elif send_password_reset_email:
        await domo_user.request_password_reset(
            domo_instance=auth.domo_instance, email=domo_user.email_address
        )

    return domo_user

# %% ../../nbs/classes/50_DomoUser.ipynb 49
@patch_to(DomoUsers)
async def upsert(
    self: DomoUsers,
    email_address: str,
    display_name: str = None,
    role_id: str = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop: int = 2,
    session: httpx.AsyncClient = None,
):
    test_valid_email(email_address)

    try:
        domo_user = await self.search_by_email(
            email = email_address,
            only_allow_one=True,
            debug_api=debug_api,
            session = session, 
            debug_num_stacks_to_drop = debug_num_stacks_to_drop + 1
        )

        if domo_user:
            user_property_ls = []
            if display_name:
                user_property_ls.append(
                    user_routes.UserProperty(
                        user_routes.UserProperty_Type.display_name, display_name
                    )
                )

            if role_id:
                user_property_ls.append(
                    user_routes.UserProperty(
                        user_routes.UserProperty_Type.role_id, role_id
                    )
                )

            if user_property_ls:
                await user_routes.update_user(
                    user_id=domo_user.id,
                    user_property_ls=user_property_ls,
                    auth=self.auth,
                    debug_api=debug_api,
                )
        return await DomoUser.get_by_id(auth=self.auth, user_id=domo_user.id)

    except SearchUser_NoResults as e:

        if not role_id:
            raise CreateUser_MissingRole(
                domo_instance=self.auth.domo_instance, email_address=email_address
            ) from e

        domo_user = await DomoUser.create(
            display_name=display_name or f"{email_address} - via dl {dt.date.today()}",
            email_address=email_address,
            role_id=role_id,
            debug_api=debug_api,
            session=session,
            auth = self.auth
        )

        await self.get()

        return domo_user

    # finally:
    #     if grant_ls:
    #         grant_ls = domo_role._valid_grant_ls(grant_ls)
    #         await domo_role.set_grants(grant_ls=grant_ls)

@patch_to(DomoUsers, cls_method=True)
async def upsert_user(
    cls: DomoUsers,
    auth: dmda.DomoAuth,
    email_address: str,
    display_name: str = None,
    role_id: str = None,
    debug_api: bool = False,
    debug_prn: bool = False,
    session: httpx.AsyncClient = None,
):
    test_valid_email(email_address)

    try:
        domo_user = await cls.by_email(
            email_ls=[email_address],
            auth=auth,
            only_allow_one=True,
            debug_api=debug_api,
        )

        if domo_user:
            user_property_ls = []
            if display_name:
                user_property_ls.append(
                    user_routes.UserProperty(
                        user_routes.UserProperty_Type.display_name, display_name
                    )
                )

            if role_id:
                user_property_ls.append(
                    user_routes.UserProperty(
                        user_routes.UserProperty_Type.role_id, role_id
                    )
                )

            if user_property_ls:
                await user_routes.update_user(
                    user_id=domo_user.id,
                    user_property_ls=user_property_ls,
                    auth=auth,
                    debug_api=debug_api,
                )
        return await DomoUser.get_by_id(auth=auth, user_id=domo_user.id)

    except SearchUser_NoResults as e:
        if debug_prn:
            print(f"No user match -- creating new user in {auth.domo_instance}")

        if not role_id:
            raise CreateUser_MissingRole(
                domo_instance=auth.domo_instance, email_address=email_address
            ) from e

        return await cls.create_user(
            auth=auth,
            display_name=display_name or f"{email_address} - via dl {dt.date.today()}",
            email_address=email_address,
            role_id=role_id,
            debug_api=debug_api,
            session=session,
        )

    # finally:
    #     if grant_ls:
    #         grant_ls = domo_role._valid_grant_ls(grant_ls)
    #         await domo_role.set_grants(grant_ls=grant_ls)



"""preferred response class for all API requests"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/client/99_DomoEntity.ipynb.

# %% auto 0
__all__ = ['NotImplemented', 'DomoEntity', 'DomoEntity_w_Lineage', 'DomoSubEntity', 'Entity_Relation']

# %% ../../nbs/client/99_DomoEntity.ipynb 2
from enum import Enum as DomoEnum


# %% ../../nbs/client/99_DomoEntity.ipynb 3
import domolibrary.client.DomoAuth as dmda

import abc
from dataclasses import dataclass, field

import httpx
from typing import Any

from nbdev.showdoc import patch_to


# %% ../../nbs/client/99_DomoEntity.ipynb 5
@patch_to(DomoEnum, cls_method = True)
def get(cls, value):
    
    for member in cls:
        if member.name.lower() == value.lower():
            return member
    
    return cls.default

@patch_to(DomoEnum, cls_method = True)
def _mising_(cls, value):
    for member in cls:
        if member.name.lower() == value.lower():
            return member
    
    return cls.default

# %% ../../nbs/client/99_DomoEntity.ipynb 9
class NotImplemented:
    
    @classmethod
    def _from_parent(cls, parent):
        return cls()


    async def get():
        raise NotImplementedError("This method should be implemented by subclasses.")

# %% ../../nbs/client/99_DomoEntity.ipynb 10
@dataclass
class DomoEntity(abc.ABC):
    auth: dmda.DomoAuth = field(repr=False)
    raw : dict = field(repr=False)

    id: str


    def __eq__(self, other):
        if isinstance(other, DomoEntity):
            return self.id == other.id
        
        return False
    

    @classmethod
    @abc.abstractmethod
    async def get_by_id(cls, auth: dmda.DomoAuth, entity_id: str):
        """
        Fetches an entity by its ID.
        """
        raise NotImplementedError("This method should be implemented by subclasses.")
    
    @abc.abstractmethod
    def display_url(self) -> str:
        """
        Returns the URL to display the entity in Domo.
        """
        raise NotImplementedError("This method should be implemented by subclasses.")


@dataclass
class DomoEntity_w_Lineage(DomoEntity):

    Lineage: Any = field(repr=False)

    def __post_init__(self):
        
        import domolibrary.classes.DomoLineage as dmdl
        
        self.Lineage = dmdl.DomoLineage._from_parent(auth = self.auth, 
                                                         parent = self)

    @classmethod
    @abc.abstractmethod
    async def _get_entity_by_id(cls, auth: dmda.DomoAuth, entity_id: str):
        """
        Fetches an entity by its ID.
        This method should be implemented by subclasses to fetch the specific entity type.
        """
        raise NotImplementedError("This method should be implemented by subclasses.")

# %% ../../nbs/client/99_DomoEntity.ipynb 11
@dataclass
class DomoSubEntity(abc.ABC):
    auth: dmda.DomoAuth = field(repr=False)
    parent : Any
    parent_id : str

    def __post_init__(self):
        if self.parent: 
            self.auth = self.parent.auth
            self.parent_id = self.parent.id

    @classmethod
    def _from_parent(cls, parent: DomoEntity):
        return cls(
            auth = parent.auth,
            parent = parent,
            parent_id = parent.id
        )
        
        

# %% ../../nbs/client/99_DomoEntity.ipynb 12
@dataclass
class Entity_Relation:
    auth: dmda.DomoAuth = field(repr=False)
    entity: Any
    relation_type: str

    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False

        return (
            self.entity.id == other.entity.id
            and self.relation_type == other.relation_type
        )

    @classmethod
    async def from_user_id(
        cls,
        user_id,
        relation_type: str,
        auth: dmda.DomoAuth,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ):

        # if entity_type == "USER":
        import domolibrary.classes.DomoUser as dmdu

        domo_user = await dmdu.DomoUser.get_by_id(
            user_id=user_id,
            auth=auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        return cls(entity=domo_user, relation_type=relation_type, auth=auth)

    @classmethod
    async def from_group_id(
        cls,
        group_id,
        relation_type,
        auth: dmda.DomoAuth,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
    ):

        import domolibrary.classes.DomoGroup as dmdg

        domo_group = await dmdg.DomoGroup.get_by_id(
            group_id=group_id,
            auth=auth,
            session=session,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop + 1,
        )

        return cls(entity=domo_group, relation_type=relation_type, auth=auth)



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoJupyter_Content.ipynb.

# %% auto 0
__all__ = ['DomoJupyter_Content']

# %% ../../nbs/classes/50_DomoJupyter_Content.ipynb 2
from ..routes.jupyter import get_workspace_auth_token_params, JupyterAPI_Error, JupyterAPI_WorkspaceStarted

# %% ../../nbs/classes/50_DomoJupyter_Content.ipynb 3
import os
import json
from dataclasses import dataclass, field
import datetime as dt

import httpx

import domolibrary.utils.DictDot as util_dd
from dateutil.parser import parse

import domolibrary.client.DomoAuth as dmda
import domolibrary.client.DomoError as dmde
import domolibrary.routes.jupyter as jupyter_routes

# %% ../../nbs/classes/50_DomoJupyter_Content.ipynb 7
@dataclass
class DomoJupyter_Content:
    name: str
    folder: str
    last_modified: dt.datetime
    file_type: str
    content: str

    auth: dmda.DomoJupyterAuth = field(repr=False)

    default_export_folder: str = "export"

    def __eq__(self, other):
        if self.__class__.__name__ != other.__class__.__name__:
            return False
        
        return self.folder == other.folder and self.name == other.name

    def __post_init__(self):
        dmda.test_is_jupyter_auth(self.auth)

        if self.folder.endswith(self.name):
            self.folder = self.folder.replace(self.name, "")

    @classmethod
    def _from_json(cls, obj: dict, auth: dmda.DomoJupyterAuth):
        dd = util_dd.DictDot(obj) if not isinstance(obj, util_dd.DictDot) else obj

        dc = cls(
            name=dd.name,
            folder=dd.path,
            last_modified=parse(dd.last_modified),
            file_type=dd.type,
            auth=auth,
            content=obj.get("content"),
        )

        return dc

    def export(
        self,
        output_folder: str = None,
        file_name: str = None,
        default_export_folder: str = None,
    ):
        if default_export_folder:
            self.default_export_folder = default_export_folder

        output_folder = output_folder or os.path.join(
            self.default_export_folder, self.folder
        )

        file_name = file_name or self.name

        if not os.path.exists(output_folder):
            print(output_folder)
            os.makedirs(output_folder)

        content_str = self.content
        if isinstance(self.content, dict):

            content_str = json.dumps(self.content)

        output_path = os.path.join(output_folder, file_name)
        with open(output_path, "w", encoding="utf-8") as f:
            f.write(content_str)
            f.close()

        return output_path
    
    @classmethod
    async def create_content(
        cls,
        auth: dmda.DomoJupyterAuth,
        new_content = None,
        folder_path = "",
        debug_api: bool = False,
        return_raw: bool = False,
        debug_num_stacks_to_drop: int = 2,
        session: httpx.AsyncClient = None,
    ):

        res = await jupyter_routes.create_jupyter_obj(
            auth=auth,
            new_content = new_content,
            content_path=folder_path,
            debug_api=debug_api,
            parent_class=cls.__name__,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            session=session,
        )

        if return_raw: 
            return res

        dj_content = cls._from_json(auth = auth, obj = res.response)

        if new_content:
            dj_content.content = new_content
            await dj_content.update()

        return dj_content

    async def update(
        self,
        jupyter_folder: str = None,
        jupyter_file_name: str = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop : int = 2,
        session : httpx.AsyncClient = None,
    ):
        if jupyter_folder and jupyter_file_name:
            content_path = f"{jupyter_folder}/{jupyter_file_name}"

        if len(self.folder) > 0:
            content_path = f"{self.folder}/{self.name}"

        else:
            content_path = self.name

            if content_path.lower().startswith(self.default_export_folder.lower()):
                content_path = content_path.replace(self.default_export_folder, "")

        content_path = "/".join(os.path.normpath(content_path).split(os.sep))

        return await jupyter_routes.update_jupyter_file(
            auth=self.auth,
            content_path=content_path,
            new_content=self.content,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session = session
        )

    async def delete(
        self,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop: int = 2,
    ):

        content_path=self.folder + '/' if self.folder else ""
        content_path += self.name

        return await jupyter_routes.delete_jupyter_content(
            auth=self.auth,
            content_path=f"{self.folder}/{self.name}",
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=self.__class__.__name__,
            session=session,
        )



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/user_attributes.ipynb.

# %% auto 0
__all__ = ['UserAttributes_IssuerType', 'UserAttributes_GET_Error', 'UserAttributes_CRUD_Error', 'get_user_attributes',
           'get_user_attribute_by_id', 'clean_attribute_id', 'generate_create_user_attribute_body',
           'create_user_attribute', 'update_user_attribute', 'delete_user_attribute']

# %% ../../nbs/routes/user_attributes.ipynb 2
from ..client.DomoEntity import DomoEnum
from typing import List

import datetime as dt

import re

import httpx
import domolibrary.client.get_data as gd
import domolibrary.client.DomoError as de

import domolibrary.client.DomoAuth as dmda

# %% ../../nbs/routes/user_attributes.ipynb 4
class UserAttributes_IssuerType(DomoEnum):
    IDP = "idp"
    SYSTEM = "domo-defined"
    CUSTOM = "customer-defined"


class UserAttributes_GET_Error(de.DomoError):
    def __init__(
        self,
        status,
        message,
        domo_instance,
        parent_class: str = None,
        function_name: str = None,
    ):
        super().__init__(
            status=status,
            message=message,
            parent_class=parent_class,
            domo_instance=domo_instance,
            function_name=function_name,
        )


class UserAttributes_CRUD_Error(de.DomoError):
    def __init__(
        self,
        status,
        message,
        domo_instance,
        parent_class: str = None,
        function_name: str = None,
    ):
        super().__init__(
            status=status,
            message=message,
            parent_class=parent_class,
            domo_instance=domo_instance,
            function_name=function_name,
        )

# %% ../../nbs/routes/user_attributes.ipynb 6
@gd.route_function
async def get_user_attributes(
    auth: dmda.DomoAuth,
    issuer_type_ls: List[
        UserAttributes_IssuerType
    ] = None,  # use `UserAttributes_IssuerType` enum
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class=None,
    debug_num_stacks_to_drop=1,
):
    """retrieves user attributes from Domo
    user attributes can be `UserAttributes_IssuerType` -- idp, domo, or user-generated
    API does not filter on the issuer type (despite API accepting filter parameter)
    """
    issuer_type_ls = issuer_type_ls or [member for member in UserAttributes_IssuerType]

    issuer_types = ",".join([member.value for member in issuer_type_ls])

    params = {"issuerTypes": issuer_types}

    url = f"https://{auth.domo_instance}.domo.com/api/user/v1/properties/meta/keys"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        params=params,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise UserAttributes_GET_Error(
            status=res.status,
            message=res.response,
            function_name=res.traceback_details.function_name,
            parent_class=parent_class,
            domo_instance=auth.domo_instance,
        )

    res.response = [
        obj
        for obj in res.response
        if obj["keyspace"] in [member.value for member in issuer_type_ls]
    ]

    return res

# %% ../../nbs/routes/user_attributes.ipynb 9
@gd.route_function
async def get_user_attribute_by_id(
    auth: dmda.DomoAuth,
    attribute_id: str,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class=None,
    debug_num_stacks_to_drop=1,
):
    """retrieves user attributes from Domo
    user attributes can be `UserAttributes_IssuerType` -- idp, domo, or user-generated
    API does not filter on the issuer type (despite API accepting filter parameter)
    """

    res = await get_user_attributes(auth=auth, session=session)

    if not res.is_success:
        raise UserAttributes_GET_Error(
            status=res.status,
            message=res.response,
            function_name=res.traceback_details.function_name,
            parent_class=parent_class,
            domo_instance=auth.domo_instance,
        )

    res.response = next(
        (obj for obj in res.response if obj["key"] == attribute_id), None
    )

    if not res.response:
        raise UserAttributes_GET_Error(
            status=res.status,
            message=f"attribute {attribute_id} not found",
            function_name=res.traceback_details.function_name,
            parent_class=parent_class,
            domo_instance=auth.domo_instance,
        )

    return res

# %% ../../nbs/routes/user_attributes.ipynb 12
def clean_attribute_id(text):
    return re.sub(r"[^A-Za-z0-9]", "", text)

# %% ../../nbs/routes/user_attributes.ipynb 13
def generate_create_user_attribute_body(
    attribute_id: str,
    name: str = None,
    description: str = None,
    issuer_type: UserAttributes_IssuerType = None,
    security_voter: str = None,
    data_type: str = None,
):
    s = {"key": attribute_id}

    if issuer_type:
        s.update({"keyspace": issuer_type.value})

    if security_voter:
        s.update({"securityVoter": security_voter})

    if data_type:
        s.update({"validator": data_type})

    if name:
        s.update({"title": name})

    if description:
        s.update({"description": description})

    return s


@gd.route_function
async def create_user_attribute(
    auth: dmda.DomoAuth,
    attribute_id,
    name=None,
    description=None,
    data_type: str = None,
    security_voter=None,
    issuer_type: UserAttributes_IssuerType = None,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class=None,
    debug_num_stacks_to_drop=1,
):
    name = name or attribute_id
    attribute_id = clean_attribute_id(attribute_id)
    description = (
        description
        or f"updated via domolibrary {dt.datetime.now().strftime('%Y-%m-%d - %H:%M')}"
    )
    data_type = data_type or "ANY_VALUE"
    security_voter = security_voter or "FULL_VIS_ADMIN_IDP"
    issuer_type = issuer_type or UserAttributes_IssuerType.CUSTOM

    body = generate_create_user_attribute_body(
        attribute_id=attribute_id,
        issuer_type=issuer_type,
        name=name,
        data_type=data_type,
        security_voter=security_voter,
        description=description,
    )

    url = f"https://{auth.domo_instance}.domo.com/api/user/v1/properties/meta/keys/{attribute_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="POST",
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise UserAttributes_CRUD_Error(
            status=res.status,
            message=f"Bad Request - does attribute {attribute_id} already exist?",
            function_name=res.traceback_details.function_name,
            parent_class=parent_class,
            domo_instance=auth.domo_instance,
        )

    res.response = f"created {attribute_id}"

    return res

# %% ../../nbs/routes/user_attributes.ipynb 15
@gd.route_function
async def update_user_attribute(
    auth: dmda.DomoAuth,
    attribute_id,
    name=None,
    description=None,
    issuer_type: UserAttributes_IssuerType = UserAttributes_IssuerType.CUSTOM,
    data_type: str = None,
    security_voter=None,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class=None,
    debug_num_stacks_to_drop=1,
):
    """body must include all attribute parameters,
    this route wil use the `get_user_attribute_by_id` function to retrieve existing values (and throw an error if not found)
    to construct an update statement"""

    body = generate_create_user_attribute_body(
        attribute_id=attribute_id,
        issuer_type=issuer_type,
        name=name,
        description=description,
        data_type=data_type,
        security_voter=security_voter,
    )

    res = await get_user_attribute_by_id(attribute_id=attribute_id, auth=auth)

    body = {**res.response, **body}

    url = f"https://{auth.domo_instance}.domo.com/api/user/v1/properties/meta/keys/{attribute_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="PUT",
        body=body,
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise UserAttributes_CRUD_Error(
            status=res.status,
            message=f"Bad Request - error updating {attribute_id}",
            function_name=res.traceback_details.function_name,
            parent_class=parent_class,
            domo_instance=auth.domo_instance,
        )

    res.response = f"updated {attribute_id}"

    return res

# %% ../../nbs/routes/user_attributes.ipynb 18
@gd.route_function
async def delete_user_attribute(
    auth: dmda.DomoAuth,
    attribute_id,
    session: httpx.AsyncClient = None,
    debug_api: bool = False,
    parent_class=None,
    debug_num_stacks_to_drop=1,
):
    url = f"https://{auth.domo_instance}.domo.com/api/user/v1/properties/meta/keys/{attribute_id}"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="DELETE",
        session=session,
        debug_api=debug_api,
        parent_class=parent_class,
        num_stacks_to_drop=debug_num_stacks_to_drop,
    )

    if not res.is_success:
        raise UserAttributes_CRUD_Error(
            status=res.status,
            message=f"Bad Request - failed to delete {attribute_id}",
            function_name=res.traceback_details.function_name,
            parent_class=parent_class,
            domo_instance=auth.domo_instance,
        )

    res.response = f"deleted {attribute_id}"

    return res



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoDataflow.ipynb.

# %% ../../nbs/classes/50_DomoDataflow.ipynb 2
from __future__ import annotations

import domolibrary.utils.DictDot as util_dd
import domolibrary.utils.chunk_execution as dmce

from domolibrary.client.DomoEntity import DomoEntity_w_Lineage
import domolibrary.client.DomoError as dmde
import domolibrary.client.DomoAuth as dmda

import domolibrary.routes.dataflow as dataflow_routes

import domolibrary.classes.DomoLineage as dmdl
import domolibrary.classes.DomoJupyter as dmdj

from dataclasses import dataclass, field
from typing import List


import httpx

from nbdev.showdoc import patch_to

# %% auto 0
__all__ = ['DomoDataflow', 'DomoDataflows']

# %% ../../nbs/classes/50_DomoDataflow.ipynb 3
from .DomoDataflow_Action import DomoDataflow_Action
from domolibrary.classes.DomoDataflow_History import (
    DomoDataflow_History,
    DomoDataflow_ActionResult,
)

# %% ../../nbs/classes/50_DomoDataflow.ipynb 6
@dataclass
class DomoDataflow(DomoEntity_w_Lineage):
    id: str
    auth: dmda.DomoAuth = field( repr=False)
    Lineage : dmdl.DomoLineage = field(repr= False)
    
    name: str = None
    owner: str = None
    description: str = None
    tags: list[str] = None
    actions: list[DomoDataflow_Action] = None

    version_id: int = None
    version_number: int = None
    versions: List[dict] = None  # list of DomoDataflow Versions

    jupyter_workspace_config : dict = None

    History: DomoDataflow_History = None  # class for managing the history of a dataflow

    JupyterWorkspace : dmdj.DomoJupyterWorkspace = None


    def __post_init__(self):
        self.History = DomoDataflow_History(
            dataflow=self, dataflow_id=self.id, auth=self.auth
        )

        self.Lineage = dmdl.DomoLineage._from_parent(
            auth = self.auth,
            parent = self
        )

    @classmethod
    def _from_json(cls, obj, auth, version_id=None, version_number=None):
        dd = util_dd.DictDot(obj)

        domo_dataflow = cls(
            auth=auth,
            id=dd.id,
            raw = obj,
            name=dd.name,
            description=dd.description,
            owner=dd.owner or dd.responsibleUserId,
            tags=dd.tags,
            version_id=version_id,
            version_number=version_number,
            Lineage = None
        )

        if dd.actions:
            domo_dataflow.actions = [
                DomoDataflow_Action._from_json(action) for action in dd.actions
            ]

            [
                domo_action.get_parents(domo_dataflow.actions)
                for domo_action in domo_dataflow.actions
            ]

        return domo_dataflow


    def display_url(self):
        return f"https://{self.auth.domo_instance}.domo.com/datacenter/dataflows/{self.id}/details"
    
    @classmethod
    async def get_by_id(
        cls: DomoDataflow,
        dataflow_id: int,
        auth: dmda.DomoAuth,
        return_raw: bool = False,
        session: httpx.AsyncClient = None,
        debug_api: bool = False,
        debug_num_stacks_to_drop=2,
        id = None
    ):
        dataflow_id = dataflow_id or id
        res = await dataflow_routes.get_dataflow_by_id(
            auth=auth,
            dataflow_id=dataflow_id,
            debug_api=debug_api,
            debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            parent_class=cls.__name__,
            session=session,
        )

        if return_raw:
            return res

        if not res.is_success:
            return None
            

        return cls._from_json(res.response, auth=auth)
    
    @classmethod
    async def _get_entity_by_id(cls, auth, entity_id, **kwargs):
        return await cls.get_by_id(
            dataflow_id=entity_id,
            auth=auth,
            return_raw=False,
            **kwargs
        )

# %% ../../nbs/classes/50_DomoDataflow.ipynb 9
@patch_to(DomoDataflow)
async def get_definition(
    self,
    debug_api: bool = False,
    return_raw: bool = False,
    session: httpx.AsyncClient = None,
):
    res = await self.get_by_id(
        auth=self.auth, 
        dataflow_id=self.id, 
        return_raw=return_raw,
        session=session,
        debug_api = debug_api,
    )

    if return_raw:
        return res


    new_obj = res.__dict__

    # Update attributes of ClassB instance
    for key, value in new_obj.items():
        if hasattr(self, key):
            setattr(self, key, value)

    return self


@patch_to(DomoDataflow)
async def update_dataflow_definition(
    self,
    new_dataflow_definition,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
):
    await dataflow_routes.update_dataflow_definition(
        auth=self.auth,
        dataflow_id = self.id,
        dataflow_definition=new_dataflow_definition,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        session=session,
    )

    return await self.get_definition(return_raw=False)

# %% ../../nbs/classes/50_DomoDataflow.ipynb 15
@patch_to(DomoDataflow)
async def get_jupyter_config(
    self,
    return_raw: bool = False,
    debug_api: bool = False,
    debug_num_stacks_to_drop : int = 2,
    session : httpx.AsyncClient = None,
):
    import domolibrary.classes.DomoJupyter as dmdj
    
    res = await dataflow_routes.search_dataflows_to_jupyter_workspaces(
        auth=self.auth,
        dataflow_id = self.id,
        debug_api = debug_api,
        debug_num_stacks_to_drop = debug_num_stacks_to_drop,
        session = session,
        parent_class = self.__class__.__name__,
        return_raw = return_raw
    )

    if return_raw:
        return res
    
    self.jupyter_workspace = await dmdj.DomoJupyterWorkspace.get_by_id(auth = self.auth, workspace_id = res.response['workspaceId'])
    
    self.jupyter_workspace_config = res.response
    self.jupyter_workspace_config['workspace_name'] = self.jupyter_workspace.name

    return self.jupyter_workspace

# %% ../../nbs/classes/50_DomoDataflow.ipynb 17
@patch_to(DomoDataflow)
async def execute(
    self: DomoDataflow,
    auth: dmda.DomoAuth = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
):
    return await dataflow_routes.execute_dataflow(
        auth=auth or self.auth,
        dataflow_id=self.id,
        parent_class=self.__class__.__name__,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
    )

# %% ../../nbs/classes/50_DomoDataflow.ipynb 21
@patch_to(DomoDataflow, cls_method=True)
async def get_by_version_id(
    cls: DomoDataflow,
    auth: dmda.DomoAuth,
    dataflow_id: int,
    version_id: int,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):
    res = await dataflow_routes.get_dataflow_by_id_and_version(
        auth=auth,
        dataflow_id=dataflow_id,
        version_id=version_id,
        debug_api=debug_api,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=cls.__name__,
        session=session,
    )

    if return_raw:
        return res

    domo_dataflow = cls._from_json(
        res.response["dataFlow"],
        version_id=res.response["id"],
        version_number=res.response["versionNumber"],
        auth=auth,
    )

    return domo_dataflow

# %% ../../nbs/classes/50_DomoDataflow.ipynb 23
@patch_to(DomoDataflow)
async def get_versions(
    self: DomoDataflow,
    debug_api: bool = False,
    debug_num_stacks_to_drop=2,
    session: httpx.AsyncClient = None,
    return_raw: bool = False,
):
    res = await dataflow_routes.get_dataflow_versions(
        auth=self.auth,
        dataflow_id=self.id,
        debug_api=debug_api,
        session=session,
        debug_num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=self.__class__.__name__,
    )

    if return_raw:
        return res

    version_ids = [df_obj["id"] for df_obj in res.response]

    self.versions = await dmce.gather_with_concurrency(
        *[
            DomoDataflow.get_by_version_id(
                dataflow_id=self.id,
                version_id=version_id,
                auth=self.auth,
                session=session,
                debug_api=debug_api,
                debug_num_stacks_to_drop=debug_num_stacks_to_drop,
            )
            for version_id in version_ids
        ],
        n=10
    )

    return self.versions

# %% ../../nbs/classes/50_DomoDataflow.ipynb 25
@dataclass
class DomoDataflows:
    auth: dmda.DomoAuth = field(repr = False)
    dataflows : List[DomoDataflow] = None

    async def get(self, return_raw: bool = False, debug_api : bool = False, session : httpx.AsyncClient = None, ):
        res = await dataflow_routes.get_dataflows(
            debug_api = debug_api,
            auth = self.auth,
            session = session,
            
        )

        if return_raw:
            return res
        
        return await dmce.gather_with_concurrency(*[
            DomoDataflow.get_by_id(auth = self.auth, dataflow_id = obj['id']) for obj in res.response], n = 10
        )
        



# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/classes/50_DomoCertification.ipynb.

# %% auto 0
__all__ = ['DomoCertificationState', 'DomoCertification']

# %% ../../nbs/classes/50_DomoCertification.ipynb 2
import datetime as dt
from dataclasses import dataclass
from domolibrary.client.DomoEntity import DomoEnum, DomoSubEntity
import domolibrary.utils.convert as cd
import domolibrary.utils.DictDot as util_dd

# %% ../../nbs/classes/50_DomoCertification.ipynb 3
class DomoCertificationState(DomoEnum):
    CERTIFIED = "certified"
    PENDING = "PENDING"
    EXPIRED = "EXPIRED"


@dataclass
class DomoCertification(DomoSubEntity):
    certification_state: DomoCertificationState
    last_updated: dt.datetime
    certification_type: str
    certification_name: str

    @classmethod
    def _from_parent(cls, parent):

        certification = parent.raw['certification']

        cert_state = None
    
        if isinstance(certification.get('state'), dict):
                cert_state = DomoCertificationState[certification.get['state'].get('value')]
    
        if isinstance(certification.get('state'), str):
              cert_state = DomoCertificationState[
                certification['state']]
                                                       
        return cls(
            auth = parent.auth,
            parent = parent,
            parent_id = parent.id,
            certification_state= cert_state,              
                                                       
            last_updated=cd.convert_epoch_millisecond_to_datetime(certification.get('lastUpdated')),
            certification_type=certification.get('processType'),
            certification_name=certification.get('processName'),
        )

    @classmethod
    def _from_json(cls,data,parent = None, parent_id = None, auth = None,  ):
        """
        Create a DomoCertification from a dictionary.
        """
        return cls(
            auth=auth,
            parent=parent,
            parent_id=parent_id,
            certification_state=DomoCertificationState[data['state']],
            last_updated=cd.convert_epoch_millisecond_to_datetime(data['lastUpdated']),
            certification_type=data['processType'],
            certification_name=data['processName'],
        )



"""Fill in a module description here"""

# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/integrations/shortcut_fn.ipynb.

# %% auto 0
__all__ = ['share_domo_account_with_domo_group']

# %% ../../nbs/integrations/shortcut_fn.ipynb 2
from typing import List
import asyncio

import domolibrary.client.DomoAuth as dmda
import domolibrary.classes.DomoGroup as dmdg
import domolibrary.classes.DomoAccount as dmacc



# %% ../../nbs/integrations/shortcut_fn.ipynb 6
async def share_domo_account_with_domo_group(
    auth: dmda.DomoAuth, 
    account_name : str,
    group_name: str,
    
    is_upsert_group : bool = False, # will not attempt to upsert group
    
    group_description: str =  None,
    group_owner_names: List[str] = None,  # will default to ["Role: Admin"]
    
    group_type: dmdg.GroupType_Enum = dmdg.GroupType_Enum["CLOSED"].value,
    
    is_search_system_accounts : bool = True # will default to True if no group_owner_name
):
    
        
    try:
        domo_groups = await dmdg.DomoGroups.by_name(auth = auth,
                                                   group_names = [group_name],
                                                   is_search_system_accounts = is_search_system_accounts)

        domo_group = next((domo_group for domo_group in domo_groups if domo_group.name == group_name))
    
    except dmdg.SearchGroups_Error as e:
        is_upsert_group = True
        
    if is_upsert_group:
        if not group_owner_names:
            is_search_system_accounts = True 
    
    
        domo_group = await dmdg.DomoGroups.upsert(auth = auth,
                                                  group_name = group_name,
                                                  description = group_description,
                                                  group_owner_names = group_owner_names,
                                                  group_type = group_type,
                                                  is_search_system_accounts = is_search_system_accounts )
    
    domo_accounts = await dmacc.DomoAccounts.get_accounts(auth = auth, 
                                                          account_name = account_name)
    
    domo_account = next((acc for acc in domo_accounts if acc.name == account_name))

    return await domo_account.share(group_id=domo_group.id, auth=auth)




# AUTOGENERATED! DO NOT EDIT! File to edit: ../../nbs/routes/auth.ipynb.

# %% auto 0
__all__ = ['AuthError', 'InvalidCredentialsError', 'AccountLockedError', 'InvalidAuthTypeError', 'InvalidInstanceError',
           'NoAccessTokenReturned', 'get_full_auth', 'get_developer_auth', 'who_am_i', 'elevate_user_otp']

# %% ../../nbs/routes/auth.ipynb 2
from typing import Optional, Any, List

import httpx

import domolibrary.client.ResponseGetData as rgd
import domolibrary.client.Logger as lg
import domolibrary.client.DomoError as dmde
import domolibrary.utils.chunk_execution as dmce

# %% ../../nbs/routes/auth.ipynb 5
class AuthError(dmde.RouteError):
    def __init__(self, **kwargs):
        super().__init__(**kwargs)


class InvalidCredentialsError(AuthError):
    """return invalid credentials sent to API"""

    def __init__(self, res, **kwargs):
        super().__init__(res=res, **kwargs)


class AccountLockedError(AuthError):
    """return invalid credentials sent to API"""

    def __init__(self, res, **kwargs):
        super().__init__(res=res, **kwargs)


class InvalidAuthTypeError(AuthError):
    """return invalid Auth type sent to API"""

    def __init__(
        self,
        res = None,
        required_auth_type: Optional[Any] = None,
        required_auth_type_ls: Optional[List[Any]] = None,
        domo_instance: Optional[str] = None,
        **kwargs
    ):

        message = f"This API rquires {required_auth_type.__name__ if required_auth_type else ', '.join([auth_type.__name__ for auth_type in required_auth_type_ls])}"

        super().__init__(message=message, res=res, domo_instance=domo_instance, **kwargs)


class InvalidInstanceError(AuthError):
    """return if invalid domo_instance sent to API"""

    def __init__(self, res, **kwargs):
        super().__init__(res=res, **kwargs)


class NoAccessTokenReturned(AuthError):
    def __init__(self, res, **kwargs):
        super().__init__(res=res, **kwargs)

# %% ../../nbs/routes/auth.ipynb 7
async def get_full_auth(
    domo_instance: str,  # domo_instance.domo.com
    domo_username: str,  # email address
    domo_password: str,
    auth: Any = None,
    session: Optional[httpx.AsyncClient] = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop=1,
    return_raw: bool = False,
) -> rgd.ResponseGetData:
    """uses username and password authentication to retrieve a full_auth access token"""

    import domolibrary.client.get_data as gd

    domo_instance = domo_instance or auth.domo_instance

    url = f"https://{domo_instance}.domo.com/api/content/v2/authentication"

    body = {
        "method": "password",
        "emailAddress": domo_username,
        "password": domo_password,
    }

    res = await gd.get_data(
        auth=auth,
        method="POST",
        url=url,
        body=body,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
        return_raw=return_raw,
    )

    if return_raw:
        return res

    if res.status == 403 and res.response == "Forbidden":
        raise InvalidInstanceError(
            res = res,
            domo_instance = domo_instance
        )

    if res.is_success and isinstance(res.response, dict):

        if res.response.get("reason") == "INVALID_CREDENTIALS":
            res.is_success = False

            raise InvalidCredentialsError(
                domo_instance = domo_instance,
                res = res

            )

        if res.response.get("reason") == "ACCOUNT_LOCKED":
            res.is_success = False

            raise AccountLockedError(
                domo_instance = domo_instance,
                res =res
            )

        if res.response == {} or res.response == "":  # no access token
            res.is_success = False

            raise NoAccessTokenReturned(
                domo_instance= domo_instance,
                res = res
            )

    if not res.response.get("sessionToken"):
        res.is_success = True

        raise InvalidCredentialsError(
            domo_instance = domo_instance,
            res = res
        )

    return res

# %% ../../nbs/routes/auth.ipynb 15
async def get_developer_auth(
    domo_client_id: str,
    domo_client_secret: str,
    auth : Any = None,
    session: Optional[httpx.AsyncClient] = None,
    debug_api: bool = False,
    parent_class: str = None,
    debug_num_stacks_to_drop= 1,
    return_raw: bool = False
) -> rgd.ResponseGetData:
    
    """
    only use for authenticating against apis documented under developer.domo.com
    """

    import domolibrary.client.get_data as gd

    url = "https://api.domo.com/oauth/token?grant_type=client_credentials"

    if not session:
        is_close_session = True
    
    # add basic auth to the session
    session = httpx.AsyncClient(
            auth=httpx.BasicAuth(domo_client_id, domo_client_secret)
        )
    

    res = await gd.get_data(method="GET",
                            url=url,
                            session = session,
                            debug_api = debug_api,
                            auth = auth,
                            num_stacks_to_drop = debug_num_stacks_to_drop,
                            parent_class = parent_class, 
                            return_raw = return_raw
                            )

    
    if is_close_session:
        await session.aclose()

    if return_raw:
        return res
     
    if res.status == 401 and res.response == "Unauthorized":
        res.is_success = False
        raise InvalidCredentialsError(
            res = res,
        )

    return res

# %% ../../nbs/routes/auth.ipynb 19
async def who_am_i(
    auth: Any,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
    debug_num_stacks_to_drop=0,
    debug_api: bool = False,
    return_raw: bool = False,
):
    """
    will attempt to validate against the 'me' API.
    This is the same authentication test the Domo Java CLI uses.
    """

    import domolibrary.client.get_data as gd

    url = f"https://{auth.domo_instance}.domo.com/api/content/v2/users/me"

    res = await gd.get_data(
        auth=auth,
        url=url,
        method="GET",
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session=session,
        return_raw=return_raw,
    )

    if return_raw:
        return res

    if res.status == 403 and res.response == "Forbidden":
        raise InvalidInstanceError(
            res = res
        )

    if res.status == 401 and res.response == "Unauthorized":
        res.is_sucess = False

        raise InvalidCredentialsError(
            res = res,
        )

    return res

# %% ../../nbs/routes/auth.ipynb 21
async def elevate_user_otp(
    auth,
    one_time_password: str,
    user_id: str = None,
    debug_api: bool = False,
    debug_num_stacks_to_drop=1,
    session: httpx.AsyncClient = None,
    parent_class: str = None,
):
    import domolibrary.client.get_data as gd

    
    if not auth.user_id and not user_id:
        await auth.who_am_i()

    user_id = user_id or auth.user_id
    
    url = f"https://{auth.domo_instance}.domo.com/api/identity/v1/authentication/elevations/{user_id}"

    body = {"timeBasedOneTimePassword": one_time_password}

    res = await gd.get_data(
        auth=auth,
        method="PUT",
        url=url,
        body=body,
        debug_api=debug_api,
        num_stacks_to_drop=debug_num_stacks_to_drop,
        parent_class=parent_class,
        session = session
    )

    if not res.is_success:
        raise InvalidCredentialsError(
            res=res
        )
    
    return res
